@@ -1,5 +1,5 @@
 # choose a mode among the following
-mode = 'simple' # 'chebyshev' # 'chebyshev_auto' # 'none'
+mode = 'linear' # options: 'linear' # 'chebyshev' # 'linear_azimuthal' # 'chebyshev_azimuthal' #'none' # 'custom'
 # not used parameters will be ignored
 order_chebyshev = 10 
 # auto masking parameters ('chebyshev_auto')
@@ -8,4 +8,7 @@ k_sigma=2.0 # this measures how much the peak stands out from the data
 peak_expand=2 # this should be about the width of the peaks
 # choose q-range where the baseline should be drawn (can be important for 'chebyshev_auto' & 'chebyshev')
 q_min=0 
-q_max=100
\ No newline at end of file
+q_max=100
+# custom mode
+path_custom_bl = 'analysis/custom_baselines.h5' # path relative to sample_dir
+h5path_custom_bl = 'baselines' # h5 internal path
\ No newline at end of file
