@@ -8,16 +8,12 @@ q_range = (10,35)
 cifPath = 'analysis/BaCO3.cif'
 # parameters for diffractlet calculation
 cutoff_structure_factor=1e-4
-max_hkl=4
+max_hkl=6
 
 odf_mode = 'hsh' # 'grid' # 
-grid_resolution = 15 # degree # ignored if hsh
-hsh_max_order = 12 # ignored if gridbased
+hsh_max_order = 12 # ignored if odf_mode is grid
+grid_resolution = 15 # degree # ignored if odf_mode is hsh
 
-# classic mode parameters
-# approximate crystal size along (a,b,c)-axes in nm
-crystalsize = (15,15,15)
-# angular sampling
-sampling = 'cubochoric' # or 'simple' (legacy)
-# angular sampling resolution
-dchi = 2*np.pi / 120
+# options
+use_structure_factors_from_data = False
+use_upgraded_point_group = True
\ No newline at end of file
