@@ -13,7 +13,7 @@ detector_direction_positive_90 = (0,0,-1) # this conforms to standard pyFAI outp
 # sample movements:
 inner_axis = (0,0,1) # inner rotation axis
 outer_axis = (0,1,0) # outer rotation axis
-scan_mode = 'line' #'column' # 'line_snake' # 'column_snake'
+scan_mode = 'line' #'column' # 'line_snake' # 'column_snake' # 'controt'
 flip_fov=False # flip fast and slow axis argument (for the case it was defined the reverse way in the integrated data files)
 
 # For calculating projectors:
