@@ -8,27 +8,24 @@ import matplotlib.pyplot as plt
 from time import time
 import sys
 
-from textom.src.model.model_textom import model_textom
 import textom.src.plotting as myplt
 import textom.src.model.rotation as rot
 import textom.src.model.symmetries as sym
+from textom.src.odf import hsh
 
 #################### input ####################
-input_file = 'input.input_simsample'
-
+gaussian_mu = np.array( (0, np.pi/2, np.pi), np.float64 ) # mean orientation of gaussian
 gaussian_std = 40 * np.pi/180 # standard deviation in rad
 nmax = 12 # maximum HSH order
-
-K = None # apply Mason kernel on the result 
-#(only necessary for low orders or sharp distributions)
+odf_resolution = 3 # degree
+symmetry = '222'
 ###############################################
 
-# mean orientation of gaussian (zero by default)
-gaussian_mu = np.array( (0, np.pi/2, np.pi), np.float64 )
-# initialize model
-mod = model_textom(input_file, single=True)
-mod.Gc[mod.Gc[:,0]==0,0] = 0.001
-gen = sym.generators(mod.symmetry)
+Gc, dV, V_fz = rot.sample_fundamental_zone( dchi=odf_resolution*np.pi/180, 
+                        sampling='cubochoric', symmetry=symmetry )
+odf = hsh.odf(nmax, sym.get_ppg_notation(symmetry), Gc)
+Gc[Gc[:,0]==0,0] = 0.001
+gen = sym.generators(symmetry)
 
 def gaussian_3d( g, mu, gen, std ):
     # make quaternion from mu
@@ -38,12 +35,11 @@ def gaussian_3d( g, mu, gen, std ):
     # for omega_mu = 0, then dg is omega
     dg = rot.ang_distance(Q_points, np.tile(q_mu,(Q_points.shape[0],1)), gen)
     odf = np.exp( - dg**2/(2*std**2) )
-    return odf/( odf @ mod.dV )
+    return odf/odf.sum()
 
 # gaussian reference-ODF
-odf_g = gaussian_3d( mod.Gc, gaussian_mu, gen, gaussian_std )
+odf_g = gaussian_3d( Gc, gaussian_mu, gen, gaussian_std )
 
-mod.set_orders( nmax )
 # initialize HSH coefficients
 c = np.zeros( mod.sHSHs.shape[0]+1 )
 c[0]=1
@@ -51,7 +47,7 @@ c_next = c.copy()
 
 # functions for optimizing HSHs against the gaussian
 def f_loss( c ):
-    odf_mod = mod.odfFromC_old( c, info=False )
+    odf_mod = odf.get_odf()
     return (( odf_g - odf_mod )**2).sum()
 
 def f_jac( c ):
