@@ -23,43 +23,46 @@ from ...config import data_type
 
 ############################################################################################+
     
-def plot_odf( orientations, odf, symmetry, num_samples=1000, title='' ):
+# def plot_odf( orientations, odf, symmetry, num_samples=1000, title='' ):
 
-    # draw samples out of the odf
-    cdf = np.cumsum( odf/odf.sum() )
+#     # draw samples out of the odf
+#     cdf = np.cumsum( odf/odf.sum() )
 
-    # Draw random samples from a uniform distribution
-    random_numbers = np.random.rand(int(num_samples))
+#     # Draw random samples from a uniform distribution
+#     random_numbers = np.random.rand(int(num_samples))
 
-    # Find the corresponding indices in the CDF
-    indices = np.searchsorted(cdf, random_numbers)
-    indices[indices==odf.size] = 0
+#     # Find the corresponding indices in the CDF
+#     indices = np.searchsorted(cdf, random_numbers)
+#     indices[indices==odf.size] = 0
 
-    # make them quaternions
-    quat = rot.QfromOTP( orientations[indices] )
-    # sample_r = Rotation([quat])
-    # sample = Orientation(sample_r)
-    # sample.symmetry = getattr( osym, sym.get_SFnotation( symmetry ) )
+#     # make them quaternions
+#     quat = rot.QfromOTP( orientations[indices] )
+#     # sample_r = Rotation([quat])
+#     # sample = Orientation(sample_r)
+#     # sample.symmetry = getattr( osym, sym.get_SFnotation( symmetry ) )
 
-    # # plots a 3D odf in the fundamental zone
-    # sample.scatter()
-    plot_points_in_fz( quat, symmetry )
-    plt.title(title)
-    plt.show()
+#     # # plots a 3D odf in the fundamental zone
+#     # sample.scatter()
+#     plot_points_in_fz( quat, symmetry )
+#     plt.title(title)
+#     plt.show()
 
 def plot_odf( unit_quaternions, odf, symmetry, num_samples=1000, title='' ):
-    # draw samples out of the odf
-    cdf = np.cumsum( odf/odf.sum() )
+    if odf.sum() == 0:
+        print('ODF normalization equals 0, provide other coefficents')
+    else:
+        # draw samples out of the odf
+        cdf = np.cumsum( odf/odf.sum() )
 
-    # Draw random samples from a uniform distribution
-    random_numbers = np.random.rand(int(num_samples))
+        # Draw random samples from a uniform distribution
+        random_numbers = np.random.rand(int(num_samples))
 
-    # Find the corresponding indices in the CDF
-    indices = np.searchsorted(cdf, random_numbers)
-    indices[indices==odf.size] = 0
+        # Find the corresponding indices in the CDF
+        indices = np.searchsorted(cdf, random_numbers)
+        indices[indices==odf.size] = 0
 
-    # # plots a 3D odf in the fundamental zone
-    plot_points_in_fz( unit_quaternions[indices], symmetry, title=title )
+        # # plots a 3D odf in the fundamental zone
+        plot_points_in_fz( unit_quaternions[indices], symmetry, title=title )
 
 def plot_points_in_fz( unit_quaterions, symmetry, title='' ):
     # make them quaternions
