@@ -33,7 +33,7 @@ def mumottize( sample_dir,
               align_by_transmission=False,
               q_index_range=(0,5), q_range=False,
               geo_path='input/geometry.py',
-              flip_fov = False, rec_1d=False, check_ori=False,
+              rec_1d=False,
               ):
     
     geo = import_module_from_path('geometry', geo_path)
@@ -43,7 +43,8 @@ def mumottize( sample_dir,
 
     dat, kappa, omega, airscat, dy = [], [], [], [], []
 
-    filelist = get_data_list(sample_dir, pattern)
+    sort_by_angles = False if geo.scan_mode == 'controt' else True
+    filelist = get_data_list(sample_dir, pattern, subdir=sub_data, sort_by_angles=sort_by_angles)
 
     # filelist = sorted( os.listdir( os.path.join(path,sub_data)) )
     # filelist = [s for s in filelist if pattern in s]#[:2]
@@ -54,27 +55,78 @@ def mumottize( sample_dir,
     #     print('\t\tDid not find angles in files, just sorted data alphabetically')
 
     if 'controt' in geo.scan_mode:
-        for k, file in enumerate( filelist ):
-            # convert q-range to indices
-            if np.any(q_range):
-                with h5py.File(os.path.join(sample_dir, sub_data, file),'r') as hf:
-                    q_in = hf['radial_units'][()]
-                    q_index_range = find_indices_in_range(q_in,q_range)
-            # continuous rotation mode
-            with h5py.File(os.path.join(sample_dir, sub_data, file),'r') as hf:
-                omega.append( hf['rot'][()] )
-                kappa.append( np.zeros_like(omega) )
-                dy.append(np.mean(hf['dty']))
+        # continuos rotation import
+        print('\tLoading data')
+        all_data = []
+        y_positions, z_positions = [], []
+        for f, file in enumerate(filelist):
+            with h5py.File(file, 'r') as hf:
+                if f==0:
+                    q_in = np.squeeze(hf['radial_units'][()] ).astype(data_type)
+                    chi_in = np.squeeze( hf['azimuthal_units'][()]*np.pi/180 ).astype(data_type)
+                    omega = ( np.array(hf['rot_angle'][()]) ).astype(data_type)
+                    kappa = np.zeros_like(omega)
+                    
+                y_positions.append( np.mean(hf['ty'][()]) )
+                z_positions.append( hf['tz'][()] )    
+
                 d = hf['cake_integ'][()]
-            if d.ndim==3:
-                dat.append( 
-                    d[:,q_index_range[0]:q_index_range[1]].sum(axis=2).mean(axis=1) )
-            elif d.ndim==2:
-                dat.append( 
-                    d[:,q_index_range[0]:q_index_range[1]].sum(axis=1) )
-            sys.stdout.write(f'\r\t\t Projection {k+1:d} / {len(filelist):d}')
-            sys.stdout.flush()
-        print('')
+                if d.ndim==3:
+                    all_data.append( 
+                        d[:,:,q_index_range[0]:q_index_range[1]].mean(axis=1) )
+                elif d.ndim==2:
+                    all_data.append( 
+                        d[:,q_index_range[0]:q_index_range[1]] )
+                    # all_data.append( ( np.array(hf['cake_integ'][()]) ).astype(data_type) )
+
+        # # find out y and z grids                   
+        y_indices = find_unique_positions(y_positions)
+        z_indices = find_unique_positions(z_positions)
+
+        n_rot_angles = omega.size
+        n_ybins = y_indices.max() + 1
+        n_zbins = z_indices.max() + 1
+        max_shape = (n_ybins, n_zbins)
+        projections = np.zeros((n_rot_angles, n_ybins, n_zbins, *all_data[0].shape[1:])) # check size and see if this would be good for preprocess_data()
+        for k,d in enumerate(all_data):
+            projections[:,y_indices[k],z_indices[k],...] = d
+        # and then the data is properly ordered and can go into textom
+            # need to y/z check scanning directions
+
+        print('\tCheck if data is displayed correctly, then close the figure.')
+        plt.figure()
+        plt.imshow(projections[0].sum(axis=2).T)
+        plt.xlabel('here should be the bottom of the sample')
+        plt.show(block=True)
+        happy = input('\t\thappy? (y/n) ')
+        if 'n' in happy:
+            sys.tracebacklimit = 0
+            raise exc.FOV('Data import not correct, check scanmode and consider enabling flip_fov = True')
+        else:
+            pass
+            # return q_in
+
+        # for k, file in enumerate( filelist ):
+        #     # convert q-range to indices
+        #     if np.any(q_range):
+        #         with h5py.File(os.path.join(sample_dir, sub_data, file),'r') as hf:
+        #             q_in = hf['radial_units'][()]
+        #             q_index_range = find_indices_in_range(q_in,q_range)
+        #     # continuous rotation mode
+        #     with h5py.File(os.path.join(sample_dir, sub_data, file),'r') as hf:
+        #         omega.append( hf['rot'][()] )
+        #         kappa.append( np.zeros_like(omega) )
+        #         dy.append(np.mean(hf['dty']))
+        #         d = hf['cake_integ'][()]
+        #     if d.ndim==3:
+        #         dat.append( 
+        #             d[:,q_index_range[0]:q_index_range[1]].sum(axis=2).mean(axis=1) )
+        #     elif d.ndim==2:
+        #         dat.append( 
+        #             d[:,q_index_range[0]:q_index_range[1]].sum(axis=1) )
+        #     sys.stdout.write(f'\r\t\t Projection {k+1:d} / {len(filelist):d}')
+        #     sys.stdout.flush()
+        # print('')
     else:
         # check if geometry is correct
         q_in = import_scan_data(sample_dir, sub_data, filelist[0],
@@ -95,16 +147,16 @@ def mumottize( sample_dir,
         airscat = np.array([p[3] for p in parallel_res], data_type)     
         print(f'\t\ttook {(time()-t1)/60} min')
 
-    if 'controt' in geo.scan_mode:
-        # sort indices according to horizontal scan
-        scan_idx = np.argsort(dy)
-        dat = np.array(dat)[scan_idx].T
-        # .T[:,scan_idx]
-        max_shape = (dat.shape[1], 1)
-        # projections = np.zeros((dat.shape[0],*max_shape, 1), data_type)
-        projections = dat.reshape((*dat.shape, 1, 1))
+    # if 'controt' in geo.scan_mode:
+    #     # sort indices according to horizontal scan
+    #     scan_idx = np.argsort(dy)
+    #     dat = np.array(dat)[scan_idx].T
+    #     # .T[:,scan_idx]
+    #     max_shape = (dat.shape[1], 1)
+    #     # projections = np.zeros((dat.shape[0],*max_shape, 1), data_type)
+    #     projections = dat.reshape((*dat.shape, 1, 1))
 
-    else:
+    # else:
         print('\tPadding Data')
         max_shape = (0,0)
         for d in dat:
@@ -130,7 +182,7 @@ def mumottize( sample_dir,
 
     if rec_1d: 
         # make a h5 file containing all the projections for all q-values
-        mm_datafile = os.path.join(sample_dir,f'analysis/rec_1d/data_rec1d.h5')
+        mm_datafile = os.path.join(sample_dir,'analysis','rec_1d','data_rec1d.h5')
         with h5py.File(mm_datafile,'w') as hf:
             hf.create_dataset( 'radial_units', data= np.array(q_in) )
             hf.create_dataset( 'projections', 
@@ -145,7 +197,7 @@ def mumottize( sample_dir,
         # pretending that the projections are the inverse of an absorbtion measurement
         diod = projections.mean(axis=3)
         diod = 1-diod/diod.max()
-        diod[diod > 1] = 1
+        diod[diod > 1] = 1.
 
         os.makedirs( os.path.join(sample_dir,'analysis/'), exist_ok=True )
         mm_datafile = os.path.join(sample_dir,f'analysis/data_mumott.h5')
@@ -169,7 +221,7 @@ def mumottize( sample_dir,
             hf.create_dataset( 'p_direction_0', data=np.array(
                 geo.beam_direction, data_type) )
             hf.create_dataset( 'volume_shape', data=np.array((max_shape[0],max_shape[0],max_shape[1]), np.int32) )
-            hf.create_dataset( 'fov_flipped', data=flip_fov)
+            hf.create_dataset( 'fov_flipped', data=geo.flip_fov)
 
             gp = hf.create_group('projections')
             for k, proj in enumerate(projections):
@@ -191,6 +243,35 @@ def get_q_index_range(path, sub_data, file,q_range):
             q_index_range = find_indices_in_range(q_in,q_range)
     return q_index_range
 
+def find_unique_positions( positions ):
+    """This assumes that positions are on a 1d grid but there is noise.
+
+    Finds out the grid points and puts them in ascending order.
+    Categorizes the positions into the grid points and returns the indices in the grid.
+
+    Parameters
+    ----------
+    positions : ndarray, float
+
+    Returns
+    ----------
+    indices: 1darray, int
+        corresponding indices for each position to the grid
+    """
+    positions = np.array(positions).flatten()
+    diffs = np.diff(np.sort(positions))
+    if diffs.max() > 5*diffs.min(): # if there is several points at the same grid point
+        threshold = np.median(diffs) + np.std(diffs)
+        step = np.median(diffs[diffs > threshold])  # ignore steps that come from noise
+        a0 = np.min(positions)
+        # Quantize each value to its nearest grid index
+        indices = np.round((np.array(positions) - a0) / step).astype(int)
+        # unique_idx = np.unique(idx)
+    else:
+        # only one z-position
+        indices = np.argsort(positions) 
+    return indices
+
 def import_scan_data(
         sample_dir, sub_data, file,
         scan_mode,
@@ -228,6 +309,14 @@ def import_scan_data(
             elif d.ndim==3:
                 d = d[:,:,q_index_range[0]:q_index_range[1]].astype(data_type).mean(axis=1)
 
+        # automatically recognize hor. and vert. scan direction
+        ty = np.atleast_1d(hf['ty'][()])
+        y_scan_direction = np.sign(ty[-1] - ty[0])
+        tz = np.atleast_1d(hf['tz'][()])
+        z_scan_direction = np.sign(tz[-1] - tz[0])
+
+        idx_data = np.arange(d.shape[0])
+
     if flip_fov:
         fov = np.flip(fov)
 
@@ -236,12 +325,14 @@ def import_scan_data(
     else:
         d = d.sum(axis=1)
         d = d.reshape(*fov,1)
+    idx_data = idx_data.reshape(fov)
 
     if 'line' in scan_mode:
         fov = np.flip( fov )
         # d = np.fliplr(np.flipud( np.transpose( d, axes=(1,0,2)) ) )
         d = np.transpose( d, axes=(1,0,2))
         # d = np.rot90(d, 2)
+        idx_data = np.transpose( idx_data )
 
     # reorder data so that they are all treated the same in the model
     if 'snake' in scan_mode:
@@ -249,6 +340,14 @@ def import_scan_data(
         for ii in range(d.shape[0]):
             if ii % 2 != 0:
                 d[ii] = d[ii][::-1]
+                idx_data[ii] = idx_data[ii][::-1]
+
+    # account for hor. and vert. scan direction
+    if y_scan_direction < 0:
+        d = np.flip(d, axis=0)
+        idx_data = np.flip(idx_data, axis=0)
+    if z_scan_direction < 0:
+        idx_data = np.flip(idx_data, axis=1)
 
     # get airscattering:
     airscat = min( d[0,0].mean(), d[0,-1].mean(), d[-1,0].mean(), d[-1,-1].mean() )
@@ -266,7 +365,7 @@ def import_scan_data(
         else:
             return q_in
     else:
-        return d, kappa, omega, airscat
+        return d, kappa, omega, airscat, idx_data
 
 #################################################### for actual alignment ####################################################
 def align( path_in, 
@@ -280,6 +379,8 @@ def align( path_in,
         output = True, ):
 
     dc = DataContainer( path_in )
+    d = dc.data[:, :, :, 0]
+    diodes_ref = np.moveaxis( d, 0, -1)
     if do_align:
         if np.any(offsets_start):
             # take given offsets as start values
@@ -328,6 +429,10 @@ def align( path_in,
             shifts = np.column_stack( (dc.geometry.j_offsets,dc.geometry.k_offsets) )
             sinogram = np.moveaxis(result['projections'][:,:,:,0], 0, -1)
             tomogram = result['reconstruction']
+
+        # shape j x k x p
+        diodes_shifted = imp.imshift_fft(diodes_ref, shifts[:, 0], shifts[:, 1])
+
     else:
         shifts = np.zeros( (len(dc.geometry.j_offsets), 2), data_type )
         fov = dc.projections[0].diode.shape
@@ -341,6 +446,9 @@ def align( path_in,
             hf.create_dataset('shifts', data=shifts)
             hf.create_dataset('sinogram', data=sinogram)
             hf.create_dataset('tomogram', data=tomogram )
+            if do_align:
+                hf.create_dataset('unshifted_data', data=diodes_ref )
+                hf.create_dataset('shifted_data', data=diodes_shifted )
             hf.create_dataset('omega', data=dc.geometry.inner_angles )
             hf.create_dataset('kappa', data=dc.geometry.outer_angles )
             hf.create_dataset('alignment_mode', data=mode)
@@ -621,7 +729,7 @@ def check_projections( path,
         output=output,
     )
 
-def backproject_all( path, path_input, range_q=None, path_data=None, only_mumottize=False, batch_size=10 ):
+def backproject_all( sample_dir, path_input, range_q=None, path_data=None, only_mumottize=False, batch_size=10 ):
     """calculates a scalar tomogram for each q-value
 
     Parameters
@@ -646,7 +754,7 @@ def backproject_all( path, path_input, range_q=None, path_data=None, only_mumott
     """
     inp = import_module_from_path('input_1Drec', path_input )
     
-    with h5py.File( os.path.join(path,'analysis/alignment_result.h5'), 'r' ) as hf:
+    with h5py.File( os.path.join(sample_dir,'analysis/alignment_result.h5'), 'r' ) as hf:
         # kap = hf['kappa'][()]
         # ome = hf['omega'][()]
         shifts = hf['shifts'][()]
@@ -656,15 +764,15 @@ def backproject_all( path, path_input, range_q=None, path_data=None, only_mumott
     # shifts=np.delete( shifts, inp.exclude_proj, axis=0 )
  
     if not path_data:
-        path_data = mumottize(path, sub_data=inp.subdir_data, pattern=inp.pattern, 
+        path_data = mumottize(sample_dir, sub_data=inp.subdir_data, pattern=inp.pattern, 
             q_index_range=[0,None],
-            geo_path=os.path.join(path,'analysis/geometry.py'),
-            flip_fov=inp.flip_fov, rec_1d=True,
+            geo_path=os.path.join(sample_dir,'analysis/geometry.py'),
+            rec_1d=True,
             )
     
     if not only_mumottize:
         # load mumott file from alignment
-        dc = DataContainer(os.path.join(path,'analysis/data_mumott.h5'))
+        dc = DataContainer(os.path.join(sample_dir,'analysis/data_mumott.h5'))
 
         #Load metadata
         with h5py.File(path_data, 'r') as hf:
@@ -678,8 +786,8 @@ def backproject_all( path, path_input, range_q=None, path_data=None, only_mumott
             range_q = [0,-1]
 
         # set up output file
-        title = os.path.basename(os.path.dirname(path))
-        out_path = os.path.join(path,'analysis/rec_1d/',
+        title = os.path.basename(os.path.dirname(sample_dir))
+        out_path = os.path.join(sample_dir,'analysis/rec_1d/',
                         f'reconstruction_1Dmm_{title}_{timestring()}.h5')
         with h5py.File( out_path, 'w' ) as hf:
             hf.create_dataset( 'sample_name', data=title )
@@ -696,7 +804,7 @@ def backproject_all( path, path_input, range_q=None, path_data=None, only_mumott
         for l in range( n_batches ):
             print('\tLoading batch')
             with h5py.File(path_data, 'r') as hf:
-                projections = hf['projections'][int(range_q[0])+l*batch_size:(l+1)*batch_size]
+                projections = hf['projections'][int(range_q[0])+l*batch_size:int(range_q[0])+(l+1)*batch_size]
             sino, tomo = [], []
             # reconstruct each q-value
             print('\tStarting reconstructions')
@@ -743,8 +851,7 @@ def backproject_qranges( path, path_input, path_data=None, only_mumottize=False
     if not path_data:
         path_data = mumottize(path, sub_data=inp.subdir_data, pattern=inp.pattern, 
             q_index_range=[np.min(inp.roi_list), np.max(inp.roi_list)],
-            geo_path=os.path.join(path,'analysis/geometry.py'),
-            flip_fov=inp.flip_fov, rec_1d=True,
+            geo_path=os.path.join(path,'analysis/geometry.py'), rec_1d=True,
             )
     
     if not only_mumottize:
