@@ -470,7 +470,7 @@ def get_file_path(package_name, filename):
         print(f"Error retrieving file path: {e}")
         return None
 
-def open_with_editor(filepath):
+def open_with_editor(filepath, confirm=True):
     """Opens the given file with the default text editor."""
     print(f"Opening file in editor: {os.path.basename(filepath)}")
     try:
@@ -482,7 +482,8 @@ def open_with_editor(filepath):
             # subprocess.run([editor, filepath])
         else:
             raise OSError(f"Unsupported OS: {os.name}")
-        input('\tPress enter when finished editing')
+        if confirm:
+            input('\tPress enter when finished editing')
     except Exception as e:
         print(f"Failed to open the file: {e}")
 
