@@ -159,11 +159,12 @@ def start_integration_parallel( sample_dir:str, confirm=True, ignore=[] ):
     print('Starting parallel pyFAI integration')
     t0 = time()
     # start parallel integration in separate processes
-    int_path = hdl.get_file_path('textom',os.path.join('src','integration_launcher.py'))
+    int_path = hdl.get_file_path('textom',os.path.join('src','integration','integration_launcher.py'))
     pids = []
     for k in range(par.n_tasks):
         command = [
-            'taskset', '-c', '%d-%d' % (k*par.cores_per_task, (k+1)*par.cores_per_task),
+            'taskset', '-c', 
+            '%d-%d' % (k*par.cores_per_task, (k+1)*par.cores_per_task-1),
             'python', int_path, 
             '-k', '%d' % (k),
             '-d', '%s' % (sample_dir),
@@ -206,13 +207,14 @@ def start_integration_online( sample_dir:str, wait=5., confirm=True, ignore=[],
     while len(filtered_datasets) > 0:
         try:
             if parallel:
-                int_path = hdl.get_file_path('textom',os.path.join('src','integration_launcher.py'))                
+                int_path = hdl.get_file_path('textom',os.path.join('src','integration','integration_launcher.py'))                
                 #for gpu just launch parallel processes that will be sent to the gpu
                 print(f'\tIntegrating {filtered_datasets[:par.n_tasks]}')
                 pids = []
                 for k in range(par.n_tasks):
                     command = [
                         'python', int_path, 
+                        '%d-%d' % (k*par.cores_per_task, (k+1)*par.cores_per_task-1),
                         '-k', f'{k}',
                         '-d', sample_dir,
                     ]
@@ -224,8 +226,10 @@ def start_integration_online( sample_dir:str, wait=5., confirm=True, ignore=[],
                     p.kill()
             else:
                 print(f'\tIntegrating {filtered_datasets[0]}')
-                flexible_integrator(sample_dir, fid_in, filtered_datasets[0], par, ai, flat, mask)
-        except:
+                flexible_integrator(sample_dir, fid_in, filtered_datasets[0], par, ai, flat, mask,
+                                    average_frames=par.average_frames)
+        except Exception as e: 
+            print('\t\t'+e)
             print(f'\t\tIntegration failed, ignoring dataset {filtered_datasets[0]}')
             ignore.append(filtered_datasets[0])
         fid_in.close()
@@ -246,7 +250,7 @@ def start_test_integration( sample_dir:str, dset_no=0 ):
     fid_in, filtered_datasets = update_filelist( sample_dir, par, ignore=[] )
     print(f'\tIntegrating {filtered_datasets[dset_no]}')
     flexible_integrator(sample_dir, fid_in, filtered_datasets[dset_no], par, ai, flat, mask,
-                        frame_average=True)
+                        pick_center=True)
     # add "_test" to filenames
     list_of_int_files = glob.glob(os.path.join(sample_dir,'data_integrated','*.h5'))
     latest_file = max(list_of_int_files, key=os.path.getctime)
@@ -263,7 +267,7 @@ def start_test_integration( sample_dir:str, dset_no=0 ):
     return test_filenames
 
 def flexible_integrator(sample_dir:str, fid_in, dataset:str, par:integration_parameters, ai, flat, mask, 
-                        frame_average=False):
+                        pick_center=False, average_frames=1):
     """Lauches pyfai integration, used in both parallel and online mode,
     both for 1d or 2d integration. Writes data and metadata in sample_dir/data_integrated/
 
@@ -283,10 +287,14 @@ def flexible_integrator(sample_dir:str, fid_in, dataset:str, par:integration_par
         flat field correction
     mask : ndarray, 2d
         raw detector mask
+    pick_center : bool
+        only integrate some data for speed
+    average_frames : int
+        made for continuous rotation mode to reduce angular sampling
     """
     # get paths for the correct h5 file
     h5path_data = f'{dataset}/{par.h5_data_path}'
-    metadata = read_metadata(fid_in, dataset, par)
+    metadata = read_metadata(fid_in, dataset, par, average_frames=average_frames)
 
     #     out_name = '{}_{:03d}_{:03d}_{:.0f}_{:08.2f}_diff_scan_0001_comb'.format(
     #                 par.title,
@@ -297,18 +305,25 @@ def flexible_integrator(sample_dir:str, fid_in, dataset:str, par:integration_par
     # else:
     out_name = dataset.split('.')[0]
 
-    data_in = fid_in[h5path_data]
-    if frame_average:
+    data_in = fid_in[h5path_data]#[()]
+    if pick_center:
         # print('yay')
         # data_in = np.atleast_3d( np.mean(
             # data_in[ metadata['npts_fast_scanaxis'] * metadata['npts_fast_scanaxis'] // 2 : metadata['npts_fast_scanaxis'] * ( metadata['npts_fast_scanaxis'] // 2 + 1 )],
             # axis=0 ) )
-        try:
+        try: #what am i trying to do here? does it work?
             data_in = data_in[ metadata['npts_slow_scanaxis'] * metadata['npts_fast_scanaxis'] // 2 : metadata['npts_slow_scanaxis'] * ( metadata['npts_fast_scanaxis'] // 2 + 1 )]
         except:
             pass
         # data_in = np.atleast_2d(np.max(data_in,axis=0))
         # print('yo')
+
+    # average data if required
+    if average_frames > 1:
+        m = data_in.shape[0] // average_frames
+        summed_frames = data_in[:m * average_frames].reshape((m, average_frames, *data_in.shape[1:])).sum(axis=1)
+        data_in = summed_frames
+
     n_frames = data_in.shape[0]
 
     if par.mode%2:
@@ -405,7 +420,7 @@ def flexible_integrator(sample_dir:str, fid_in, dataset:str, par:integration_par
         fid_out.close()
         print(f'Total time: {time()-t0:.1f} s')
 
-def read_metadata( fid_in, dataset, par:integration_parameters ):
+def read_metadata( fid_in, dataset, par:integration_parameters, average_frames=1 ):
     """Reads metadata from raw file as given in the integration_parameters.py file
 
     Parameters
@@ -435,8 +450,6 @@ def read_metadata( fid_in, dataset, par:integration_parameters ):
         ('tilt_angle', par.h5_tilt_angle_path),
         ('ty', par.h5_ty_path),
         ('tz', par.h5_tz_path),
-        ('npts_fast_scanaxis',par.h5_nfast_path),
-        ('npts_slow_scanaxis',par.h5_nslow_path),
     ]
     for dset_name, dset_path in required:
         try:
@@ -458,20 +471,25 @@ def read_metadata( fid_in, dataset, par:integration_parameters ):
                 raise exc.metadata_missing(f'Required metadata in {h5path} not found')
 
     # this differs depending on scan or controt
+        # ('npts_fast_scanaxis',par.h5_nfast_path),
+        # ('npts_slow_scanaxis',par.h5_nslow_path),
     if par.h5_nfast_path:
         try:
             h5path_nfast = f'{dataset}/{par.h5_nfast_path}'
             nfast = fid_in[h5path_nfast][()]
-        except:
-            print(f'\t\tRequired metadata in {h5path_nfast} not found, stopping' )
-            raise exc.metadata_missing(f'Required metadata in {h5path_nfast} not found')
-        try:
+        # except:
+        #     print(f'\t\tRequired metadata in {h5path_nfast} not found, stopping' )
+        #     raise exc.metadata_missing(f'Required metadata in {h5path_nfast} not found')
+        # try:
             h5path_nslow = f'{dataset}/{par.h5_nslow_path}'
             nslow = fid_in[h5path_nslow][()]
         except:
-            print(f'\t\tRequired metadata in {h5path_nslow} not found, stopping' )
-            raise exc.metadata_missing(f'Required metadata in {h5path_nslow} not found')
+            # print(f'\t\tRequired metadata in {h5path_nslow} not found, stopping' )
+            # raise exc.metadata_missing(f'Required metadata in {h5path_nslow} not found')
+            nfast, nslow = -1, -1
         metadata['fov'] = ( nfast, nslow )
+        metadata['npts_fast_scanaxis'] = nfast
+        metadata['npts_slow_scanaxis'] = nslow
 
     # read optional metadata
     optional = [
@@ -489,6 +507,13 @@ def read_metadata( fid_in, dataset, par:integration_parameters ):
         except:
             print(f'\t\tOptional metadata in {h5path} not found, continuing' )
 
+    # average frames if required
+    for dset_name in metadata.keys():
+        if len(np.atleast_1d(metadata[dset_name])) > 1 and average_frames > 1:
+            m = len(metadata[dset_name]) // average_frames
+            averaged = metadata[dset_name][:m * average_frames].reshape(m, average_frames).mean(axis=1)
+            metadata[dset_name] = averaged
+            
     return metadata
 
 def write_metadata( metadata, fid_out ):
