@@ -11,7 +11,7 @@ plt.ion()
 """ Library for creating masks """
 ####################################################
 class RegionSelector:
-    def __init__(self, ax, ymin, ymax, max_regions=None):
+    def __init__(self, ax, ymin, ymax, max_regions=None, peaks=None):
         """
         Initialize the RegionSelector class.
 
@@ -28,6 +28,7 @@ class RegionSelector:
         self.press = None
         self.rects = []
         self.current_rect = None
+        self.peaks = peaks # this is to put another color if no peak is contained
 
         self.cidpress = ax.figure.canvas.mpl_connect('button_press_event', self.on_press)
         self.cidrelease = ax.figure.canvas.mpl_connect('button_release_event', self.on_release)
@@ -63,9 +64,18 @@ class RegionSelector:
         release_x = event.xdata
         left = min(self.press_x,release_x)
         width = np.abs(self.press_x - release_x)
+
+        if np.any(self.peaks):
+            if np.any(np.logical_and(self.peaks>left, self.peaks<left+width)):
+                col = 'yellow'
+            else:
+                col = 'red'
+        else:
+            col= 'yellow'
+
         self.current_rect = Rectangle((left, self.ymin),
                                       width, self.ymax - self.ymin,
-                                      facecolor='yellow', edgecolor='yellow', alpha=0.5)
+                                      facecolor=col, edgecolor='magenta', alpha=0.3)
         self.ax.add_patch(self.current_rect)
         self.rects.append(self.current_rect)
 
@@ -133,7 +143,7 @@ def select_regions(data_x, data_y, data_x2=None, data_y2=None, hkl=None,
     ax.set_title(title)
     ax.legend()
     selector = RegionSelector(ax, minval, maxval, 
-                    max_regions=max_regions)
+                    max_regions=max_regions, peaks=data_x2)
     plt.show(block=True)  # Keeps the plot open until it is manually closed
     return selector#.get_regions()
 
@@ -210,7 +220,7 @@ def select_threshold(data):
     return selector.get_value()
 
 def select_threshold_hist(data, xlabel='Intensity', ylabel='Number of Voxels', 
-                          title='Click to set the lower threshold', logscale=False):
+                          title='Click to set the lower threshold', logx=False, logy=False):
     """
     Plot a histogram and allow the user to draw a vertical line by clicking in the window.
 
@@ -222,17 +232,27 @@ def select_threshold_hist(data, xlabel='Intensity', ylabel='Number of Voxels',
     """
     plt.ion()
     fig, ax = plt.subplots(figsize=(10, 6))
-    ax.hist(data, bins=50, label='1D Data')
+    nbins = 50
+    if logx:
+        bins = np.geomspace(data[data>0].min(), data.max(), nbins+1)
+        # bins = np.logspace(np.log10(data[data>0].min()),np.log10(data.max()), nbins)
+    else:
+        bins=nbins
+
+    ax.hist(data, bins=bins, label='1D Data')
     ax.set_xlabel(xlabel)
     ax.set_ylabel(ylabel)
     ax.set_title(title)
-    if logscale:
+    if logx:
+        ax.set_xscale('log')
+    if logy:
         ax.set_yscale('log')
     selector = LineSelector(ax, data, vertical=True)
     # plt.legend()
     plt.show()
     plt.ioff()
     plt.show(block=True)  # Keeps the plot open until it is manually closed
+    plt.ion()
     return selector.get_value()
 
 def mask_voxels( Kappa, Omega, ty, tz, shifty, shiftz, x_p, tomogram, sinogram ):
@@ -241,7 +261,7 @@ def mask_voxels( Kappa, Omega, ty, tz, shifty, shiftz, x_p, tomogram, sinogram )
     to_try = np.array([[0,0],[0,np.pi/2]])
     rgl = []
     for b in to_try:
-        A = np.column_stack((Kappa, Omega))
+        A = np.column_stack((Kappa, Omega % 2*np.pi))
         # check which of the projections is closest to the angles defined in totry
         distances = np.linalg.norm(A - b, axis=1)
         g_match = np.argmin(distances)
@@ -280,12 +300,23 @@ def mask_voxels( Kappa, Omega, ty, tz, shifty, shiftz, x_p, tomogram, sinogram )
             tomo_flat > cutoff_low,
             vmask_0))
 
-        # erosion dilation of noise
+        # # erosion dilation of noise
         # structure_size = 2
         # structuring_element = ndimage.generate_binary_structure(3, 1)
         # structuring_element = ndimage.iterate_structure(
         #     structuring_element, structure_size)
-        vmask = ndimage.binary_opening(vmask)#, structure=structuring_element)
+        # vmask = ndimage.binary_opening(vmask, structure=structuring_element)
+
+        # smooth surface with a filter (and grow slightly)
+        structure = ndimage.generate_binary_structure(3, 1)
+        filled = ndimage.binary_fill_holes(vmask.reshape(tomogram.shape), structure=structure)
+        smoothed = ndimage.gaussian_filter(filled.astype(float), sigma=1) > 0.4
+        # keep largest component
+        labeled, n = ndimage.label(smoothed)
+        sizes = ndimage.sum(smoothed, labeled, range(1, n + 1))
+        largest = labeled == (np.argmax(sizes) + 1)
+        vmask = largest.flatten()
+        print(f'\tSmoothed surface. New number of voxels: {vmask.sum()}')
 
         # print('\tMasking empty voxels. Choose regions to exclude in figure.')
         # sphere_mask = msk.sphere_exclusion(x_p[vmask])
@@ -325,7 +356,7 @@ def check_tomogram( tomogram, mask ):
     ax.set_title("Surface of the masked object. Verify and close figure")
 
     set_axes_equal(ax)
-    plt.show()
+    plt.show(block=True)
 
 def mask_detector( image, mask ):
     """Creates a mask on a 2D image by user-input
@@ -405,9 +436,10 @@ def mask_detector( image, mask ):
 
     plt.figure()
     plt.imshow(mask)
-    plt.show(block=True)
+    plt.title('Dark regions are masked')
+    plt.show()
 
-    return mask.flatten()
+    return mask#.flatten()
 
 ####################################################
 class RectangleDrawer:
