@@ -3,13 +3,13 @@ import os
 import sys
 import shutil
 import importlib
-from numba import njit
 import inspect
 from datetime import datetime
 import sys
 import time
 import re
 import glob
+import platform
 
 from .handle import get_file_path
 from ..version import __version__
@@ -66,53 +66,9 @@ def read_phrase():
         lines = fid.readlines()
     return lines[np.random.randint(len(lines))][:-1]
 
-# Numba-compatible functions
-@njit()
-def sparsemult(sparsearray, indices, vector):
-    """ Vector product using one custom sparse-array
+def check_version():
 
-    Parameters
-    ----------
-    sparsearray : ndarray, float
-        list of the non-zero entries
-    indices : ndarray, int
-        indices of the non-zero entries
-    vector : ndarray, float
-        conventional vector
-    
-    returns: 
-    ------------
-    sum : float
-    """
-    sum = 0
-    for i,S in zip(indices, sparsearray):
-        sum += vector[i] * S
-    return sum
-
-@njit()
-def integrate_c(sparsearray, indices, C ):
-    """ Sparse-multiplication for several entries at once 
-
-    Parameters
-    ----------
-    sparsearray : ndarray, float
-        list of the non-zero entries
-    indices : ndarray, int
-        indices of the non-zero entries
-    C : 2D ndarray, float
-        conventional array
-        dim 0: to be summed over, 1: whatever
-    
-    returns: 
-    ------------
-    c_int : ndarray float
-        vector product for every entry in dim 1 of C
-    """
-    c_int = np.zeros( C.shape[1], C.dtype)
-    for i in range(indices.shape[0]):
-        idx = indices[i]
-        c_int += C[idx] * sparsearray[i] # get weighted average hsh-coefficients
-    return c_int
+    return 0
 
 def cp_addno(source_file, destination_folder):
     """ copies a file to destination. if a file with the same
@@ -398,6 +354,36 @@ def find_indices_in_range(array, value_range):
 
     return lower_index, upper_index
 
+def find_unique_positions( positions ):
+    """This assumes that positions are on a 1d grid but there is noise.
+
+    Finds out the grid points and puts them in ascending order.
+    Categorizes the positions into the grid points and returns the indices in the grid.
+
+    Parameters
+    ----------
+    positions : ndarray, float
+
+    Returns
+    ----------
+    indices: 1darray, int
+        corresponding indices for each position to the grid
+    """
+    positions = np.array(positions).flatten()
+    diffs = np.diff(np.sort(positions))
+    if diffs.max() > 5*diffs.min(): # if there is several points at the same grid point
+        threshold = np.median(diffs) + np.std(diffs)
+        step = np.median(diffs[diffs > threshold])  # ignore steps that come from noise
+        a0 = np.min(positions)
+        # Quantize each value to its nearest grid index
+        indices = np.round((np.array(positions) - a0) / step).astype(int)
+        # unique_idx = np.unique(idx)
+    else:
+        # only one z-position
+        indices = np.argsort(positions) 
+    return indices
+
+
 def list_files_by_modification_time(directory, reverse=False, extension=None):
     """
     List all files in a directory ordered by modification time, optionally filtered by file extension.
@@ -449,4 +435,22 @@ def is_running_in_jupyter():
             return True
     except ImportError:
         pass
-    return False
\ No newline at end of file
+    return False
+
+def get_affinity_info():
+    """this is to check if the process is runnin under taskset with restricted CPUs
+    """
+    if platform.system() == "Linux":
+        allowed = sorted(os.sched_getaffinity(0))
+        return {
+            "restricted": len(allowed) < os.cpu_count(),
+            "allowed_cores": allowed,
+            "n_cores": len(allowed)
+        }
+    else:
+        # macOS / Windows → no affinity API (no taskset)
+        return {
+            "restricted": False,
+            "allowed_cores": None,
+            "n_cores": os.cpu_count()
+        }
\ No newline at end of file
