@@ -10,8 +10,7 @@ from . import symmetries as sym
 from ..odf import hsh
 from ..odf import gridbased as grd
 from .. import numba_plugins as nb
-from ..analysis import orix_plugins as orx
-from ...config import data_type
+from ...config import data_type, hex_notation, odf_resolution
 
 def parse_cif(file_path):
     """
@@ -30,6 +29,12 @@ def parse_cif(file_path):
     lattice_vectors = []
     symmetry_operations = []
     positions_cartesian = []
+    occupancy = []
+    B = []
+
+    def parse_number(s):
+        # this makes sure we can read numbers with given uncertainty, e.g. '4.9614(3)'
+        return float(s.split('(')[0])
 
     try:
         with open(file_path, 'r') as f:
@@ -43,17 +48,17 @@ def parse_cif(file_path):
             
             # Extract lattice vectors
             if line.startswith('_cell_length_a'):
-                a = float(line.split()[-1])
+                a = parse_number(line.split()[-1])
             elif line.startswith('_cell_length_b'):
-                b = float(line.split()[-1])
+                b = parse_number(line.split()[-1])
             elif line.startswith('_cell_length_c'):
-                c = float(line.split()[-1])
+                c = parse_number(line.split()[-1])
             elif line.startswith('_cell_angle_alpha'):
-                alpha = float(line.split()[-1])
+                alpha = parse_number(line.split()[-1])
             elif line.startswith('_cell_angle_beta'):
-                beta = float(line.split()[-1])
+                beta = parse_number(line.split()[-1])
             elif line.startswith('_cell_angle_gamma'):
-                gamma = float(line.split()[-1])
+                gamma = parse_number(line.split()[-1])
 
             # Find atomic data section
             if line.strip().startswith('loop_'):
@@ -72,9 +77,12 @@ def parse_cif(file_path):
                     while j < len(lines) and not lines[j].strip().startswith('loop_') and lines[j].strip():
                         split_line = lines[j].split()
                         atom_labels.append(split_line[headers.index('_atom_site_type_symbol')])  # Assume first column is atom type
-                        atom_fractions.append([float(split_line[headers.index('_atom_site_fract_x')]),
-                                            float(split_line[headers.index('_atom_site_fract_y')]),
-                                            float(split_line[headers.index('_atom_site_fract_z')])])
+                        atom_fractions.append([parse_number(split_line[headers.index('_atom_site_fract_x')]),
+                                            parse_number(split_line[headers.index('_atom_site_fract_y')]),
+                                            parse_number(split_line[headers.index('_atom_site_fract_z')])])
+                        occupancy.append(parse_number(split_line[headers.index('_atom_site_occupancy')]))
+
+                        # B.append(float(split_line[headers.index('_atom_site_U_iso_or_equiv')]))
                         j += 1
 
         # Convert lattice parameters to lattice vectors
@@ -103,14 +111,17 @@ def parse_cif(file_path):
 
         # Apply symmetry operations to fractional coordinates
         atom_fractions = np.array(atom_fractions)
-        full_fractions, atom_list = [], []
+        full_fractions, atom_list, occupancy_list, B_list = [], [], [], []
         for op in symmetry_operations:
-            for (frac, at) in zip(atom_fractions, atom_labels):
+            # for (frac, at, oc, b) in zip(atom_fractions, atom_labels, occupancy, B):
+            for (frac, at, oc) in zip(atom_fractions, atom_labels, occupancy):
                 full_fractions.append(apply_symmetry_operation(op, frac))
                 atom_list.append(at)
+                occupancy_list.append(oc)
+                # B_list.append(b)
         full_fractions = np.array(full_fractions)
         full_fractions = np.mod(full_fractions, 1)  # Ensure all coordinates are within [0, 1)
-        full_fractions, indices = np.unique(np.squeeze(full_fractions), axis=0, return_index=True)
+        full_fractions, indices = np.unique(np.squeeze(np.round(full_fractions,5)), axis=0, return_index=True)
 
         # Convert fractional to Cartesian coordinates
         positions_cartesian = np.array(
@@ -129,12 +140,20 @@ def parse_cif(file_path):
             'atom_list': np.array(atom_list)[indices],
             'fractional_positions': full_fractions,
             'cartesian_positions': positions_cartesian,
+            'occupancy': np.atleast_1d(occupancy_list)[indices],
+            # 'B': np.atleast_1d(B_list)[indices],
         }
 
     except Exception as e:
         print(f"Error parsing CIF file: {e}")
         return None
 
+def wavelength_from_E_keV(E_keV):
+    h = 4.135667696e-18 # keV*s
+    c = 299792458 # m/s
+    wavelength = h*c*1e9 / E_keV # nm
+    return wavelength
+
 def get_reciprocal_space_coordinates(Qq_det, Chi_det, E_keV, geo):
     """Calculates the coordinates of the detector points based on
     polar coordinates q_det, chi_det and X-ray energy E_kev
@@ -145,9 +164,7 @@ def get_reciprocal_space_coordinates(Qq_det, Chi_det, E_keV, geo):
         3D reciprocal space coordinates and number of detector points
     """
 
-    h = 4.135667696e-18 # keV*s
-    c = 299792458 # m/s
-    wavelength = h*c*1e9 / E_keV # nm
+    wavelength = wavelength_from_E_keV(E_keV)
 
     Two_theta = 2 * np.arcsin(Qq_det * wavelength / (4*np.pi))
     QX,QY,QZ = np.empty_like(Two_theta), np.empty_like(Two_theta), np.empty_like(Two_theta)
@@ -297,7 +314,7 @@ def group_equivalent_reflections(hkl_max, lattice_vectors, point_group_quaternio
     # Reciprocal lattice vectors
     V = np.dot(lattice_vectors[0], np.cross(lattice_vectors[1], lattice_vectors[2]))  # Unit cell volume
     reciprocal_lattice_vectors = np.column_stack([
-         2 * np.pi * np.cross(lattice_vectors[1], lattice_vectors[2]) / V,
+        2 * np.pi * np.cross(lattice_vectors[1], lattice_vectors[2]) / V,
         2 * np.pi * np.cross(lattice_vectors[2], lattice_vectors[0]) / V,
         2 * np.pi * np.cross(lattice_vectors[0], lattice_vectors[1]) / V
         ])
@@ -384,11 +401,11 @@ def equivalents_of_hkl(hkl, B, rotations, include_inversion=True,
     return unique_hkls
 
 # def plot_powder_pattern(reciprocal_points, energy_keV, num_bins=100):
-#     """Compute and plot the powder diffraction pattern (Intensity vs 2θ)."""
-#     wavelength = 12.398 / energy_keV  # Convert energy to wavelength (Å)
+#     """Compute and plot the powder diffraction pattern (Intensity vs 2??)."""
+#     wavelength = 12.398 / energy_keV  # Convert energy to wavelength (Å?)
 #     G_magnitudes = np.linalg.norm(reciprocal_points, axis=1)  # Compute |G|
     
-#     # Convert |G| to 2θ using Bragg's Law: 2θ = 2 * arcsin(λG / 4π)
+#     # Convert |G| to 2?? using Bragg's Law: 2?? = 2 * arcsin(??G / 4??)
 #     theta_2 = 2 * np.arcsin(wavelength * G_magnitudes / (4 * np.pi)) * (180 / np.pi)  # Convert to degrees
     
 #     # Create histogram (powder diffraction pattern)
@@ -398,7 +415,7 @@ def equivalents_of_hkl(hkl, B, rotations, include_inversion=True,
 #     # Plot powder pattern
 #     plt.figure(figsize=(8, 5))
 #     plt.plot(bin_centers, hist, '-k', lw=1.5)
-#     plt.xlabel('2θ (degrees)')
+#     plt.xlabel('2?? (degrees)')
 #     plt.ylabel('Intensity (arb. units)')
 #     plt.title('Simulated Powder Diffraction Pattern')
 #     plt.grid(True)
@@ -413,6 +430,64 @@ def equivalents_of_hkl(hkl, B, rotations, include_inversion=True,
 # intensity_scaling_factors = intensity_scaling_factors/intensity_scaling_factors[0]
 #####
 
+# def structure_factor(hkl, lattice_vectors, frac_coords, species,
+#                      form_factor_fn, occupancy=None, B=None):
+#     """
+#     hkl: tuple/list (h,k,l)
+#     lattice_vectors: 3x3 array-like, columns = a, b, c (in angstroms)
+#     frac_coords: Nx3 array of fractional coordinates (x,y,z)
+#     species: list of length N with element symbols matching form_factor_fn
+#     form_factor_fn: callable f(element, s) -> scalar real (atomic form factor)
+#                     where s = sin(theta)/lambda
+#     occupancy: optional list/array length N (default 1)
+#     B: optional list/array length N of isotropic B factors (angstrom^2) (default 0)
+#     Returns complex F and intensity |F|^2
+#     """
+#     h, k, l = hkl
+#     A = np.asarray(lattice_vectors).reshape(3,3)
+#     # compute G vector (reciprocal space)
+#     A_inv = np.linalg.inv(A)
+#     G = 2*np.pi * (A_inv.T @ np.array([h,k,l]))   # cartesian reciprocal vector
+#     s = np.linalg.norm(G) / (4*np.pi)             # sin(theta)/lambda
+#     coords = np.asarray(frac_coords)
+#     N = coords.shape[0]
+#     occ = np.ones(N) if occupancy is None else np.asarray(occupancy)
+#     Barr = np.zeros(N) if B is None else np.asarray(B)
+#     F = 0+0j
+#     for j in range(N):
+#         x,y,z = coords[j]
+#         f = form_factor_fn(species[j], s)
+#         phase = np.exp(2j*np.pi*(h*x + k*y + l*z))
+#         thermal = np.exp(-Barr[j]*s*s)
+#         F += occ[j] * f * thermal * phase
+#     I = (F.real**2 + F.imag**2)
+#     return F, I
+
+def structure_factor_q(hkl, lattice_vectors, frac_coords,
+                       form_factors, occupancy=None, B=None):
+    """
+    
+    """
+    h, k, l = hkl
+    A = np.asarray(lattice_vectors).reshape(3,3)
+    A_inv = np.linalg.inv(A)
+    G = 2*np.pi * (A_inv.T @ np.array([h,k,l]))
+    q = np.linalg.norm(G)
+    coords = np.asarray(frac_coords)
+    N = coords.shape[0]
+    occ = np.ones(N) if occupancy is None else np.asarray(occupancy)
+    Barr = np.zeros(N) if B is None else np.asarray(B)
+    F = 0+0j
+    for j in range(N):
+        x,y,z = coords[j]
+        f = form_factors[j]
+        thermal = np.exp(-Barr[j]*(q/(4*np.pi))**2)
+        phase = np.exp(2j*np.pi*(h*x + k*y + l*z))
+        F += occ[j] * f * thermal * phase
+    I = (F.real**2 + F.imag**2)
+
+    return F, I
+
 # @njit       
 def structure_factor( reciprocal_points, atomic_form_factors, atom_positions, bool_pos_el ):
     '''
@@ -433,25 +508,40 @@ def structure_factor( reciprocal_points, atomic_form_factors, atom_positions, bo
         structure factor of the unit cell, function of Q
         dim0: complex structure factor
     '''
-    SU_complex=np.zeros_like(reciprocal_points[:,0], np.complex128)
+    S_complex=np.zeros_like(reciprocal_points[:,0], np.complex128)
     nEl=bool_pos_el.shape[0]
     
     for i in range(nEl): #for all elements 
-        f=atomic_form_factors[i] #get atomic form factor (for all Qs)
-        sComplex=np.zeros_like(SU_complex)
+        # f=atomic_form_factors[i] #get atomic form factor (for all Qs)
+        S_element=np.zeros_like(S_complex)
         for r in atom_positions[bool_pos_el[i,:]]: #all atom positions of that element
             qr = np.dot(reciprocal_points,r)
-            sComplex += np.exp(qr*1j)
-        SU_complex += f*sComplex
-    return np.real(SU_complex*SU_complex.conjugate())
+            S_element += np.exp(qr*1j)
+        S_complex += atomic_form_factors[i] * S_element
+    return np.real(S_complex*S_complex.conjugate())
 
-def structure_factor_from_cif(cif_path, cutoff_structure_factor=1e-4, max_hkl=4, q_min=0, q_max=60, powder=False):
+def structure_factor_from_cif(cif_path, 
+            cutoff_structure_factor=1e-4, max_hkl=4, q_min=0, q_max=60, wavelength=None, powder=False, upgrade_pointgroup=False):
 
     crystal_data = parse_cif(cif_path)
     lattice_vectors = crystal_data['lattice_vectors']/10 # assumes angstroem and converts to nm
+    chem = crystal_data['atom_list'] # chemical symbols of all atoms in the unit cell
+    atom_positions = crystal_data['cartesian_positions']/10 # coordinates of all atoms in the unit cell converted to nm
+    space_group_no = crystal_data['space_group']
+    # # alternatively import with ase:
+    # import ase
+    # unit = ase.io.read(cif_path)
+    # atom_positions1 = unit.positions/10 # coordinates of all atoms converted to nm
+    # chem = np.array( unit.get_chemical_symbols() ) # chemical symbols of all atoms
+    # lattice_vectors = np.array( unit.get_cell() )/10 # converted to nm
+    # space_group_no = unit.info['spacegroup'].no
+    #
+    if upgrade_pointgroup:
+        symmetry = sym.get_diffraction_laue_group(space_group_no)
+    else:
+        symmetry = sym.get_proper_point_group(space_group_no)
 
     #### Calculate reciprocal lattice points
-    symmetry = sym.get_proper_point_group(crystal_data['space_group'])
     gen = sym.generators(symmetry)
     Q_gen = rot.QfromOTP(gen)
     Q_group = rot.generate_group(Q_gen) # point group rotational symmetries
@@ -467,9 +557,6 @@ def structure_factor_from_cif(cif_path, cutoff_structure_factor=1e-4, max_hkl=4,
     q_abs = q_abs_full[mask]
 
     #### Calculate atomic form factors
-    #
-    chem = crystal_data['atom_list'] # chemical symbols of all atoms in the unit cell
-    #
     elements = np.unique(chem)
     Nel = elements.size
     #bool mask for the unit cell
@@ -477,32 +564,41 @@ def structure_factor_from_cif(cif_path, cutoff_structure_factor=1e-4, max_hkl=4,
     for k in range(Nel):
         # determinate the positions of element k 
         atom_types_bool[k,:] = chem == elements[k]
-    #
+    # load Cromer??Mann coefficients
     ff_path = hdl.get_file_path('textom',
         os.path.join('ressources','atomic_formfactors.txt'))
     ffcoeff = np.genfromtxt( ff_path, dtype="|U8", skip_header=2 )
-    ffcoeff_used = np.array([ ffcoeff[ np.where( el == ffcoeff )[0][0], 1: ] for el in elements]).astype(data_type)
+    ffcoeff_used = np.atleast_2d([ ffcoeff[ np.where( el == ffcoeff )[0][0], 1: ] for el in elements]).astype(data_type)
     a = ffcoeff_used[:,0:-1:2]
     b = ffcoeff_used[:,1:-1:2]
     c = ffcoeff_used[:,-1]
     FF_element = np.empty( (Nel, q_abs.size), np.float64 ) # atomic form factors for all used elements
     for k in range(Nel):
-        A, Qf = np.meshgrid( a[k], q_abs ) # using angstroems^-1 here for simplicity
-        B, _ = np.meshgrid( b[k], q_abs )
+        A, Qf = np.meshgrid( a[k], q_abs/10 ) # using angstroems^-1 here for simplicity
+        B, _ = np.meshgrid( b[k], q_abs/10 )
         C = c[k]
         #Calculate form factor http://lampx.tugraz.at/~hadley/ss1/crystaldiffraction/atomicformfactors/formfactors.php
         F = ( A * np.exp( -B * (Qf/(4*np.pi))**2  ) ).sum(axis=1) + C
         # _, FF = np.meshgrid( self.Chi_det.reshape(self.detShape)[0], F )
         FF_element[k,:] = F
     #
-    # # plot to verify form factors:
+    # plot to verify form factors:
     # idcs = np.argsort(q_abs)
     # plt.plot(q_abs[idcs],FF_element[0,idcs])
     # #
 
-    #### Calculate structure factor from unit cell
-    atom_positions = crystal_data['cartesian_positions']/10 # coordinates of all atoms in the unit cell converted to nm
-    structure_factors = structure_factor( reciprocal_points, FF_element, atom_positions, atom_types_bool)
+    ### Calculate structure factor from unit cell
+    form_factors = np.array([FF_element[np.where(e==elements)[0][0]] for e in chem])
+    structure_factors = np.empty_like(form_factors[0])
+    for k in range(hkl_list.shape[0]):
+        F,I=structure_factor_q(
+            hkl_list[k], 
+            crystal_data['lattice_vectors'].T, # note that it takes input in angstroem
+            crystal_data['fractional_positions'], 
+            form_factors[:,k], crystal_data['occupancy'], None )#crystal_data['B'])
+        structure_factors[k] = I
+
+    # structure_factors = structure_factor( reciprocal_points, FF_element, atom_positions, atom_types_bool)
     # # plot to verify structure factors:
     # idcs = np.argsort(q_abs)
     # plt.plot(q_abs[idcs],structure_factors[idcs], 'x')
@@ -530,85 +626,15 @@ def structure_factor_from_cif(cif_path, cutoff_structure_factor=1e-4, max_hkl=4,
         get_reciprocal_points(lattice_vectors, np.array(list(hkl_eq))) for hkl_eq in hkl_equivalents_used]
     )
 
-    if symmetry[0] == '6': # use m
-        hkl_used = miller_to_hex(hkl_used)
+    if hex_notation and symmetry[0] == '6':
+        hkl_used = sym.miller_to_hex(hkl_used)
 
     if powder:
-        return q_used, hkl_used, structure_factors_used, multiplicities_used
+        return q_used, hkl_used, structure_factors_used, multiplicities_used, symmetry
     else:
         return q_used, reciprocal_points_used_full, structure_factors_used, multiplicities_used, hkl_used, symmetry
     # schreit nach einem structure_dict oder object
 
-def miller_to_hex(indices):
-    """
-    Convert 3-index Miller indices (h,k,l) to 4-index hexagonal (h,k,i,l).
-    
-    Parameters
-    ----------
-    indices : array-like, shape (3,) or (N,3)
-        Input Miller indices (h,k,l).
-    
-    Returns
-    -------
-    hex_indices : ndarray, shape (4,) or (N,4)
-        Converted hexagonal indices (h,k,i,l).
-    """
-    arr = np.atleast_2d(indices).astype(int)
-    h, k, l = arr[:, 0], arr[:, 1], arr[:, 2]
-    i = -(h + k)
-    hex_arr = np.stack([h, k, i, l], axis=1)
-    return hex_arr if indices.ndim > 1 else hex_arr[0]
-
-
-def hex_to_miller(indices, check=True):
-    """
-    Convert 4-index hexagonal (h,k,i,l) to 3-index Miller (h,k,l).
-    
-    Parameters
-    ----------
-    indices : array-like, shape (4,) or (N,4)
-        Input hexagonal indices (h,k,i,l).
-    check : bool
-        If True, enforce i = -(h+k).
-    
-    Returns
-    -------
-    miller_indices : ndarray, shape (3,) or (N,3)
-        Converted Miller indices (h,k,l).
-    """
-    arr = np.atleast_2d(indices).astype(int)
-    h, k, i, l = arr[:, 0], arr[:, 1], arr[:, 2], arr[:, 3]
-
-    if check:
-        expected = -(h + k)
-        if not np.all(i == expected):
-            raise ValueError(f"Inconsistent indices: expected i={expected}, got {i}")
-
-    miller_arr = np.stack([h, k, l], axis=1)
-    return miller_arr if indices.ndim > 1 else miller_arr[0]
-
-def plot_powder_pattern(cif_path, cutoff_structure_factor=1e-4, max_hkl=4, q_min=0, q_max=60, q_dat=None, I_dat=None):
-    
-    q, hkl, S, M = structure_factor_from_cif(cif_path, cutoff_structure_factor, max_hkl,  q_min=q_min, q_max=q_max, powder=True)
-    powder_intensities = S * M
-    # plt.plot( q_used, powder_intensities, 'x' )
-    # print('hkl\tq\tInt\t\tRelative I\tMultiplicity')
-    print('hkl\tq\tRelative I\tMultiplicity')
-    plt.figure()
-    for (xi, yi, l, m) in zip(q, powder_intensities, hkl, M):
-        plt.plot([xi,xi],[0,yi],'r')
-        plt.text(xi, yi, l, va='bottom', ha='center')
-        # print(f'{l}\t{xi:5.3}\t{yi:8.1f}\t{100*yi/powder_intensities.max():.1f}\t{m}')
-        print(f'{l}\t{xi:5.3}\t{100*yi/powder_intensities.max():.1f}\t\t{m}')
-    if np.any(q_dat):
-        msk_q = np.logical_and(q_dat>q_min, q_dat<q_max)
-        q_dat = q_dat[msk_q]
-        I_dat = I_dat[msk_q] * powder_intensities.max() / I_dat[msk_q].max()
-        plt.plot(q_dat, I_dat, label='Data')
-        plt.legend()
-    plt.ylim(bottom=0)
-    plt.show(block=True)
-
 # @njit
 # def quaternion_to_match_vectors(u, v):
 #     """
@@ -633,7 +659,7 @@ def plot_powder_pattern(cif_path, cutoff_structure_factor=1e-4, max_hkl=4, q_min
 #             orthogonal = np.array([0.0, 1.0, 0.0])
 #         axis = np.cross(u, orthogonal)
 #         axis /= np.linalg.norm(axis)
-#         return np.array([0.0, *axis])  # 180° rotation: w=0
+#         return np.array([0.0, *axis])  # 180°? rotation: w=0
 #     else:
 #         axis = np.cross(u, v)
 #         axis /= np.linalg.norm(axis)
@@ -752,7 +778,7 @@ def get_probability(q_peak, q_detector, odf_basis_function, odf_args, sampling=3
     Probabilities = odf_basis_function( Rotations, *odf_args )
     # plt.plot(np.linspace(-180,180,num=Probabilities.size),Probabilities)
 
-    probability_integrated = nb.nb_trapz( Probabilities, 2*np.pi/sampling )
+    probability_integrated = nb.trapz( Probabilities, 2*np.pi/sampling )
     return probability_integrated
 
 @njit
@@ -920,25 +946,37 @@ def get_diffractlets( cr, chi_det, geo, Q_sample_rotations, hsh_max_order=4, gri
     """
 
     q_hkl, reciprocal_points_full, structure_factors, multiplicities, hkl, symmetry = structure_factor_from_cif( 
-        cr.cifPath, q_max=cr.q_range[1], cutoff_structure_factor=cutoff_structure_factor, max_hkl=max_hkl )
+        cr.cifPath, q_max=cr.q_range[1], cutoff_structure_factor=cutoff_structure_factor, max_hkl=max_hkl,
+        upgrade_pointgroup=cr.use_upgraded_point_group )
+    
+    try:
+        if cr.use_structure_factors_from_data:
+            structure_factors = np.ones_like(structure_factors)
+    except:
+        print('\tCalculated structure_factors')
+
     Qq_det, Chi_det, detShape = rot.qchi(q_hkl, chi_det)
     detector_coordinates = get_reciprocal_space_coordinates( Qq_det, Chi_det, cr.E_keV, geo )
     detector_coordinates = detector_coordinates.reshape((*detShape,3))
     # pg = getattr( orx.osym, sym.get_SFnotation( symmetry ) )
     powder_pattern = structure_factors * multiplicities
-
-    print('\tRetrieved reciprocal space coordinates, calculating diffractlets')
+    Gc, dV, V_fz = rot.sample_fundamental_zone( dchi=odf_resolution*np.pi/180, 
+                            sampling='cubochoric', symmetry=symmetry )
+    
+    print('\tRetrieved reciprocal space coordinates')
     t0 = time()
     if odf_mode=='hsh':
-        ns = hsh.get_orders(symmetry, hsh_max_order, info=False)
+        odf_obj = hsh.odf(hsh_max_order, sym.get_ppg_notation(symmetry), Gc)
+        # ns = hsh.get_orders(symmetry, hsh_max_order, info=False)
+        print('\tCalculating diffractlets')
         difflets_n = []
-        for order in ns:
+        for order in odf_obj.orders:
             if order==0:
                 difflets = np.array(np.full( (Q_sample_rotations.shape[0],1,*detShape), 
                                 np.atleast_2d( structure_factors*multiplicities) ), data_type)
             else:
-                symmetrization_matrix, _,_ = hsh.get_symmetrization_matrix( np.atleast_1d(ns), symmetry )
-                difflets,_ = diffractlets_parallel_hsh( order, symmetrization_matrix[str(order)],
+                # symmetrization_matrix, _,_ = hsh.get_symmetrization_matrix( np.atleast_1d(ns), symmetry )
+                difflets,_ = diffractlets_parallel_hsh( order, odf_obj.symmetrization_matrix[str(order)],
                             detector_coordinates, Q_sample_rotations,
                             reciprocal_points_full, multiplicities, structure_factors)
                 # n_difflets = symmetrization_matrix[str(order)].shape[0]
@@ -986,8 +1024,14 @@ def get_diffractlets( cr, chi_det, geo, Q_sample_rotations, hsh_max_order=4, gri
         print('\n\tMode not recognized, choose hsh or grid\n')
         sys.exit(1)
     
+    # apply Lorentz correction
+    theta = np.arcsin(wavelength_from_E_keV(cr.E_keV) * Qq_det / (4*np.pi))
+    Lorentz = 1/np.sin(2*theta)
+    difflets_full = difflets_full * Lorentz.reshape(detShape)[np.newaxis,np.newaxis,:,:]
+
     print(f'\t\ttook {(time()-t0)/60:.2f} min')
-    return Qq_det.reshape(detShape), Chi_det.reshape(detShape), detector_coordinates, hkl, difflets_full, powder_pattern, symmetry
+    return Qq_det.reshape(detShape), Chi_det.reshape(detShape), detector_coordinates, \
+                    hkl, structure_factors, difflets_full, powder_pattern, symmetry, odf_obj
 
 def plot_diffractlet( Qq_det, Chi_det, hkl, difflet, q_bins=50, cmap='bwr', logscale=False, sym_cmap=True ):
 
@@ -1117,14 +1161,35 @@ def calculate_experimental_odf_coverage(cr, geo,
 
     stop=1
 
+@njit(parallel=True)
+def inversion_matrix_parallel(chosen_peaks, chi_det, Q_samplerot, Q_bins,
+                              detector_coordinates, reciprocal_points_full, multiplicities,
+                              Q_group):
+    n_data_proj = chosen_peaks.size * chi_det.size
+    n_proj = Q_samplerot.shape[0]
+    n_parameters = Q_bins.shape[0]
+    A = np.zeros( (n_proj, n_data_proj, n_parameters), np.int8 )
+    for r in prange(Q_samplerot.shape[0]):
+        quat_rot = Q_samplerot[r]
+        k = 0
+        for q in chosen_peaks:
+            for q_d in detector_coordinates[:,q]:
+                q_d_rot = rot.quaternion_inverse_rotate_vector( quat_rot, q_d )
+                for q_p in reciprocal_points_full[multiplicities[:q].sum():multiplicities[:q+1].sum()]:
+                    ori_detected = get_orientations_seen_on_pixel(q_d_rot, q_p, sampling=100)
+                    idcs = get_odf_bin_indices(ori_detected, Q_bins, Q_group)
+                    A[r, k, np.unique(idcs)] += 1
+                k+=1
+    return A
 
-def calculate_inversion_matrix( cr, geo, d_rot=10, tilts=[0,15,30,45], chi_resolution=6, cutoff_structure_factor=1e-4):
+def get_inversion_matrix( cr, geo, chosen_peaks, d_rot=10, tilts=[0,15,30,45], 
+                               chi_resolution=6, cutoff_structure_factor=1e-4):
 
     # let's see for a single ODF
     q_hkl, reciprocal_points_full, structure_factors, multiplicities, hkl, symmetry = structure_factor_from_cif( 
         cr.cifPath, q_max=cr.q_range[1], cutoff_structure_factor=cutoff_structure_factor )
 
-    chosen_peaks = [0,1] # check indices: hkl
+    # chosen_peaks = [0,1] # check indices: hkl
 
     Gc, dV, V_fz = rot.sample_fundamental_zone( dchi=chi_resolution*np.pi/180, 
                                         sampling='cubochoric', symmetry=symmetry )  
@@ -1146,36 +1211,43 @@ def calculate_inversion_matrix( cr, geo, d_rot=10, tilts=[0,15,30,45], chi_resol
     q_gen = rot.QfromOTP(sym.generators(symmetry))
     Q_group = rot.generate_group(q_gen)
 
-    n_data_points = len(chosen_peaks) * chi_det.size * Q_samplerot.shape[0]
-    n_parameters = Gc.shape[0]
-    A = np.zeros( (n_data_points, n_parameters), np.int8 )
-    k = 0
-    for quat_rot in Q_samplerot:
-        for q in chosen_peaks:
-            for q_d in detector_coordinates[:,q]:
-                q_d_rot = rot.quaternion_inverse_rotate_vector( quat_rot, q_d )
-                for q_p in reciprocal_points_full[multiplicities[:q].sum():multiplicities[:q+1].sum()]:
-                    ori_detected = get_orientations_seen_on_pixel(q_d_rot, q_p, sampling=100)
-                    idcs = get_odf_bin_indices(ori_detected, Q_bins, Q_group)
-                    A[k, np.unique(idcs)] += 1
+    A = inversion_matrix_parallel(np.array(chosen_peaks), chi_det, Q_samplerot, Q_bins,
+                              detector_coordinates, reciprocal_points_full, multiplicities//2,
+                              Q_group)
 
-                k+=1
+    return np.concatenate(A)
 
-    U, s, Vt = np.linalg.svd(A)
+    # n_data_points = len(chosen_peaks) * chi_det.size * Q_samplerot.shape[0]
+    # n_parameters = Gc.shape[0]
+    # A = np.zeros( (n_data_points, n_parameters), np.int8 )
+    # k = 0
+    # for quat_rot in Q_samplerot:
+    #     for q in chosen_peaks:
+    #         for q_d in detector_coordinates[:,q]:
+    #             q_d_rot = rot.quaternion_inverse_rotate_vector( quat_rot, q_d )
+    #             for q_p in reciprocal_points_full[multiplicities[:q].sum():multiplicities[:q+1].sum()]:
+    #                 ori_detected = get_orientations_seen_on_pixel(q_d_rot, q_p, sampling=100)
+    #                 idcs = get_odf_bin_indices(ori_detected, Q_bins, Q_group)
+    #                 A[k, np.unique(idcs)] += 1
 
-    # Small singular values -> unstable directions
-    tol = 1e-12  # or relative threshold like s[0]*1e-6
-    unstable = np.where(s < tol)[0]
+    #             k+=1
 
-    # Look at rows of Vt corresponding to unstable singular values
-    print("Unstable directions in coefficient space:")
-    print(Vt[unstable])
+    # U, s, Vt = np.linalg.svd(A)
+
+    # # Small singular values -> unstable directions
+    # tol = 1e-12  # or relative threshold like s[0]*1e-6
+    # unstable = np.where(s < tol)[0]
+
+    # # Look at rows of Vt corresponding to unstable singular values
+    # print("Unstable directions in coefficient space:")
+    # print(Vt[unstable])
     
-    # The ratio (largest ÷ smallest) is the condition number. 
-    # If this number is huge, your inversion will be numerically unstable: 
-    # tiny changes or noise in d will blow up in the solution c
-    condition_number = np.max(s) / np.min(s)
+    # # The ratio (largest ÷? smallest) is the condition number. 
+    # # If this number is huge, your inversion will be numerically unstable: 
+    # # tiny changes or noise in d will blow up in the solution c
+    # condition_number = np.max(s) / np.min(s)
     
     # The smallest singular value tells you how close A is to being rank-deficient (losing information). 
-    # If a singular value is exactly zero, that means some directions in coefficient space can’t be 
-    # recovered from d — the system is underdetermined.
+    # If a singular value is exactly zero, that means some directions in coefficient space can??t be 
+    # recovered from d ?? the system is underdetermined.
+
