@@ -1,6 +1,7 @@
 import numpy as np
 from scipy import special as sp
 from numba import njit, prange
+from numba.typed import List
 import math
 import os, sys
 
@@ -8,74 +9,191 @@ from .hsh_symmetrization import HSH_symmetrization_matrix, R
 from .. import handle as hdl
 from ...config import data_type
 
-def get_odf( coefficients, G_odf, symmetry, n_max ):
+class odf:
+    def __init__(self, n_max, symmetry, orientation_grid, exclude_ghosts=True):
+        """Initializes a model to calculate Orientation Distribution Functions from 
+        HyperSpherical Harmonics
 
-    nlm, shsh_basis,_,_,_ = get_symmetrized_hshs(G_odf, n_max, symmetry)
-            
-    return coefficients @ shsh_basis
+        Parameters
+        ----------
+        n_max : int
+            highest order of the harmonics expansion
+        symmetry : str
+            proper point group in notation
+        orientation_grid : ndarray, shape(n,3)
+            array of orientations in OTP notation
+        exclude_ghosts : bool
+            if true, skips the half-uneven orders (2,6,10,..)
+        """
+        print('\tInitializing hyperspherical harmonics model')
+        self.odf_mode = 'hsh'
+        self.n_max = n_max
+        self.symmetry = symmetry
+        self.orders = get_orders( symmetry, n_max, info=True, exclude_ghosts=exclude_ghosts )
+        self.G_odf = orientation_grid
+        # Properties to be initialized when needed:
+        self._basis_functions = None
+        self._symmetrization_matrix = None
+        self._slices_hsh = None
+        self._slices_shsh = None
 
-def get_odf_batch( coefficients_batch, G_odf, symmetry, n_max ):
+    @property
+    def slices_shsh( self ):
+        if self._slices_shsh is None:
+            # Number sHSHs for each n
+            n_sHSHs  =  np.array( [get_NsHSH(self.symmetry,n) for n in self.orders] ) 
+            self._slices_shsh = np.array(
+                [ [ n_sHSHs[:k].sum(), n_sHSHs[:k+1].sum() ] 
+                    for k in range(n_sHSHs.shape[0]) ]) # 'slices' of coefficients as numbers for each n
+        return self._slices_shsh
+
+    @property
+    def slices_hsh( self ):
+        if self._slices_hsh is None:
+            n_HSHs = np.array([(n+1)**2 for n in self.orders])
+            self._slices_hsh = np.array(
+                [ [ n_HSHs[:k].sum(), n_HSHs[:k+1].sum() ] 
+                    for k in range(n_HSHs.shape[0]) ]) # 'slices' of coefficients as numbers for each n
+        return self._slices_hsh
     
-    nlm, shsh_basis,_,_,_ = get_symmetrized_hshs(G_odf, n_max, symmetry)
+    @property
+    def symmetrization_matrix( self ):
+        if self._symmetrization_matrix is None:
+            # calculate the matrix that transforms HSHs to sHSHs
+            # Xsn = HSH_symmetrization_matrix(np.max(orders), symmetry)
+            self._symmetrization_matrix = {'0': np.array([[1.]], np.complex128)}
+            for n in self.orders[1:]:
+                _, csym = cSymmHSH(self.symmetry, n) # get sSHSs and orders
+                self._symmetrization_matrix[str(n)] = csym
 
-    odf_batch = odf_batch_parallel(coefficients_batch.astype(data_type), shsh_basis.astype(data_type))
+            # X_new = HSH_symmetrization_matrix(self.n_max, self.symmetry)
+            # stop=1
+            # self.symmetrization_matrix, self.slices_hsh, self.slices_shsh
+        return self._symmetrization_matrix
 
-    # odf_batch = np.empty( (coefficients_batch.shape[0], G_odf.shape[0]) )
-    # for c, coefficients in enumerate(coefficients_batch):
-    #     odf_batch[c] = coefficients @ shsh_basis
+    @property
+    def basis_functions(self):
+        if self._basis_functions is None:
+            print('\t\tCalculating sHSH basis functions')
+            # first calculate the HSHs, then symmetrize, then sum over coefficients
+            # n_hsh = self.slices_hsh[-1,1]
+            # n_shsh = self.slices_shsh[-1,1]
 
-    return odf_batch
+            n_hsh = np.diff( self.slices_hsh ).flatten()
+            n_shsh = np.diff( self.slices_shsh ).flatten()
 
-def get_odf_centered( coefficients, G_odf, symmetry, n_max ):
-    orders = get_orders( symmetry, n_max, info=False )
-    nlm, shsh_basis,_,symmetrization_matrix, slices = get_symmetrized_hshs(G_odf, n_max, symmetry)
-    odf = coefficients @ shsh_basis
-    g_mu = G_odf[ np.argmax(odf) ]
+            self._basis_functions = np.empty((n_shsh.sum(), self.G_odf.shape[0]))
+            for i_n, n in enumerate(self.orders):
+                lm = np.empty((n_hsh[i_n], 2), dtype=np.int64)
+                idx = 0
+                for l in range(n+1):
+                    for m in range(-l, l+1):
+                        lm[idx, 0] = l
+                        lm[idx, 1] = m
+                        idx += 1
+                
+                self._basis_functions[self.slices_shsh[i_n,0]:self.slices_shsh[i_n,1]] = compute_basis_functions(
+                            self.G_odf, n, lm, 
+                            self.symmetrization_matrix[str(n)], 
+                        )
 
-    c_centered = np.empty_like(coefficients)
-    c_centered[0] = coefficients[0]
-    for n in range(1,orders.size):
-        order = orders[n]
-        R_centering = Rs_n( np.array([-g_mu[0], g_mu[1], g_mu[2]], np.float64),
-                                    order, symmetrization_matrix[str(order)] )
-        c_centered[slices[n,0]:slices[n,1]] = R_centering @ coefficients[slices[n,0]:slices[n,1]]
+            # hsh_basis = Z_all_parallel(self.G_odf, nlm)
+            
+            # self._basis_functions = np.empty((n_shsh, self.G_odf.shape[0]))
+            # for i_n, n_str  in enumerate(self.symmetrization_matrix.keys()):
+            #     self._basis_functions[self.slices_shsh[i_n,0]:self.slices_shsh[i_n,1]] = np.real(
+            #         self.symmetrization_matrix[n_str] @ hsh_basis[self.slices_hsh[i_n,0]:self.slices_hsh[i_n,1]])
+                
 
-    return c_centered @ shsh_basis
+        return self._basis_functions
 
-def get_odf_centered_batch( coefficients_batch, G_odf, G_mu, symmetry, n_max ):
-    orders = get_orders( symmetry, n_max, info=False )
-    nlm, shsh_basis,_,symmetrization_matrix,slices = get_symmetrized_hshs(G_odf, n_max, symmetry)
+    def get_odf( self, coefficients ):
+                
+        return coefficients @ self.basis_functions
 
-    C_centered = np.empty_like(coefficients_batch)
-    C_centered[:,0] = coefficients_batch[:,0]
-    for n in range(1,orders.size):
-        order = orders[n]
-        C_centered[:, slices[n,0]:slices[n,1]] = rotate_coefficients_n_parallel(
-            np.column_stack( (-G_mu[:,0], G_mu[:,1], G_mu[:,2]) ).astype(np.float64),
-            coefficients_batch[:,slices[n,0]:slices[n,1]].astype(np.float64), order, 
-            symmetrization_matrix[str(order)])
+    def get_odf_batch( self, coefficients_batch ):
+        
+        odf_batch = odf_batch_parallel(coefficients_batch.astype(data_type), self.basis_functions.astype(data_type))
+
+        # odf_batch = np.empty( (coefficients_batch.shape[0], G_odf.shape[0]) )
+        # for c, coefficients in enumerate(coefficients_batch):
+        #     odf_batch[c] = coefficients @ shsh_basis
+
+        return odf_batch
+
+    def get_odf_maxima( self, coefficients_batch ):
+        i_max = odf_maximum_indices(coefficients_batch.astype(data_type), self.basis_functions.astype(data_type))
+        G_max = self.G_odf[i_max]
+        return G_max
+
+    def get_odf_centered( self, coefficients ):
+
+        odf = coefficients @ self.basis_functions
+        g_mu = self.G_odf[ np.argmax(odf) ]
+
+        c_centered = np.empty_like(coefficients)
+        c_centered[0] = coefficients[0]
+        for n in range(1,self.orders.size):
+            order = self.orders[n]
+            R_centering = Rs_n( np.array([-g_mu[0], g_mu[1], g_mu[2]], np.float64),
+                                        order, self.symmetrization_matrix[str(order)] )
+            c_centered[self.slices_shsh[n,0]:self.slices_shsh[n,1]] = R_centering @ coefficients[self.slices_shsh[n,0]:self.slices_shsh[n,1]]
+
+        return c_centered @ self.basis_functions
+
+    def get_odf_centered_batch( self, coefficients_batch, mean_orientations ):
+
+        C_centered = np.empty_like(coefficients_batch)
+        C_centered[:,0] = coefficients_batch[:,0]
+        for n in range(1,self.orders.size):
+            order = self.orders[n]
+            C_centered[:, self.slices_shsh[n,0]:self.slices_shsh[n,1]] = rotate_coefficients_n_parallel(
+                np.column_stack( (-mean_orientations[:,0], mean_orientations[:,1], mean_orientations[:,2]) ).astype(np.float64),
+                coefficients_batch[:,self.slices_shsh[n,0]:self.slices_shsh[n,1]].astype(np.float64), order, 
+                self.symmetrization_matrix[str(order)])
 
 
-        # R_centering = Rs_n_stack( np.column_stack( (-G_mu[:,0], G_mu[:,1], G_mu[:,2]) ).astype(np.float64),
-        #                             order, symmetrization_matrix[str(order)] )
-        # for v in range(C_centered.shape[0]):
-        #     C_centered[v, slices[n,0]:slices[n,1]] = R_centering[v] @ coefficients_batch[v,slices[n,0]:slices[n,1]]
+            # R_centering = Rs_n_stack( np.column_stack( (-G_mu[:,0], G_mu[:,1], G_mu[:,2]) ).astype(np.float64),
+            #                             order, symmetrization_matrix[str(order)] )
+            # for v in range(C_centered.shape[0]):
+            #     C_centered[v, slices[n,0]:slices[n,1]] = R_centering[v] @ coefficients_batch[v,slices[n,0]:slices[n,1]]
 
-    odf_batch = odf_batch_parallel(C_centered.astype(data_type), shsh_basis.astype(data_type))
+        odf_batch = odf_batch_parallel(C_centered.astype(data_type), self.basis_functions.astype(data_type))
 
-    # odf_batch = np.empty( (coefficients_batch.shape[0], G_odf.shape[0]) )
-    # for c, coefficients in enumerate(C_centered):
-    #     odf_batch[c] = coefficients @ shsh_basis
+        # odf_batch = np.empty( (coefficients_batch.shape[0], G_odf.shape[0]) )
+        # for c, coefficients in enumerate(C_centered):
+        #     odf_batch[c] = coefficients @ shsh_basis
 
-    return odf_batch
+        return odf_batch
+    
+    def get_odf_std_batch( self, coefficients_batch, mean_orientations ):
+        C_centered = np.empty_like(coefficients_batch)
+        C_centered[:,0] = coefficients_batch[:,0]
+        for n in range(1,self.orders.size):
+            order = self.orders[n]
+            C_centered[:, self.slices_shsh[n,0]:self.slices_shsh[n,1]] = rotate_coefficients_n_parallel(
+                np.column_stack( (-mean_orientations[:,0], mean_orientations[:,1], mean_orientations[:,2]) ).astype(np.float64),
+                coefficients_batch[:,self.slices_shsh[n,0]:self.slices_shsh[n,1]].astype(np.float64), order, 
+                self.symmetrization_matrix[str(order)])
+        
+        omega = self.G_odf[:,0]
+        std_batch = std_parallel( C_centered.astype(data_type), self.basis_functions.astype(data_type), omega )
+        return std_batch
 
 @njit(parallel=True)
 def odf_batch_parallel( coefficients_batch, shsh_basis ):
-    odf_batch = np.empty((coefficients_batch.shape[0],shsh_basis.shape[1]))
+    odf_batch = np.empty((coefficients_batch.shape[0],shsh_basis.shape[1]), coefficients_batch.dtype)
     for c in prange(coefficients_batch.shape[0]):
         odf_batch[c] = coefficients_batch[c] @ shsh_basis
     return odf_batch
 
+@njit(parallel=True)
+def odf_maximum_indices( coefficients_batch, shsh_basis ):
+    maximum_indices = np.empty((coefficients_batch.shape[0]), np.int32)
+    for c in prange(coefficients_batch.shape[0]):
+        maximum_indices[c] = np.argmax(coefficients_batch[c] @ shsh_basis)
+    return maximum_indices
+
 @njit(parallel=True)
 def rotate_coefficients_n_parallel(G_rot, C, order, symmetrization_matrix):
     C_rotated = np.empty_like(C)
@@ -84,34 +202,13 @@ def rotate_coefficients_n_parallel(G_rot, C, order, symmetrization_matrix):
         C_rotated[v] = R_centering @ C[v]
     return C_rotated
 
-def get_symmetrized_hshs(G_odf, n_max, symmetry):
-    # first calculate the HSHs, then symmetrize, then sum over coefficients
-    orders = get_orders( symmetry, n_max, info=False )
-
-    symmetrization_matrix, slices_hsh, slices_shsh = get_symmetrization_matrix(orders, symmetry)
-    
-    n_hsh = slices_hsh[-1,1]
-    n_shsh = slices_shsh[-1,1]
-
-    nlm = np.empty((n_hsh, 3), dtype=np.int64)
-    idx = 0
-    for n in orders:
-        for l in range(n+1):
-            for m in range(-l, l+1):
-                nlm[idx, 0] = n
-                nlm[idx, 1] = l
-                nlm[idx, 2] = m
-                idx += 1
-
-    hsh_basis = Z_all_parallel(G_odf, nlm)
-    
-    shsh_basis = np.empty((n_shsh, G_odf.shape[0]))
-    for i_n, n_str  in enumerate(symmetrization_matrix.keys()):
-        shsh_basis[slices_shsh[i_n,0]:slices_shsh[i_n,1]] = np.real(
-            symmetrization_matrix[n_str] @ hsh_basis[slices_hsh[i_n,0]:slices_hsh[i_n,1]])
-            ## why are they not purely real??
-
-    return nlm, shsh_basis, hsh_basis, symmetrization_matrix, slices_shsh
+@njit(parallel=True)
+def std_parallel( centered_coefficients_batch, shsh_basis, omega ):
+    Std = np.empty( centered_coefficients_batch.shape[0], centered_coefficients_batch.dtype )
+    for o in prange(centered_coefficients_batch.shape[0]):
+        odf_o = centered_coefficients_batch[c] @ shsh_basis
+        Std[o] = np.sqrt( ( odf_o * omega**2 ).sum() / odf_o.sum() )
+    return Std
 
 def mason_kernel(c, ns, slices, K):
     """Applies the Mason kernel on the HSH coefficients
@@ -229,15 +326,56 @@ def Z( omega,theta,phi, n,l,m ):
 # print(f'maximum deviation: {np.max(np.abs(old_Z - new_Z))}') # was about 1e-15
 ###
 
+
+
+# @njit(parallel=True)
+# def compute_basis_functions(orientations, nlm, sym_mats, slices_hsh, slices_shsh):
+#     n_shsh = slices_shsh.shape[0]
+#     n_orient = orientations.shape[0]
+#     G = np.empty((slices_shsh[-1,1], n_orient))  # final result
+
+#     for i_n in prange(n_shsh):
+#         n1, n2 = slices_hsh[i_n]
+#         s1, s2 = slices_shsh[i_n]
+#         M = sym_mats[i_n]   # e.g., prepacked array of symmetry matrices
+
+#         # build local hsh_basis
+#         nh = n2 - n1
+#         hsh_basis_local = np.empty((nh, n_orient), np.complex128)
+#         for z in range(nh):
+#             n_idx = n1 + z
+#             for i in range(n_orient):
+#                 hsh_basis_local[z, i] = Z_numba(
+#                     orientations[i, 0], orientations[i, 1], orientations[i, 2],
+#                     nlm[n_idx, 0], nlm[n_idx, 1], nlm[n_idx, 2]
+#                 )
+
+#         # multiply and store
+#         G[s1:s2] = np.real(M @ hsh_basis_local)
+#     return G
+
 @njit(parallel=True)
-def Z_all_parallel(orientations, nlm):
-    n_hsh = nlm.shape[0]
-    hsh_basis = np.empty((n_hsh, orientations.shape[0]), np.complex128)
+def compute_basis_functions(G_odf, n, lm, symm_matrix_n):
+    n_shsh, n_hsh = symm_matrix_n.shape
+    shsh_basis_n = np.zeros((n_shsh, G_odf.shape[0]), np.float64)
     for z in prange(n_hsh):
-        for i in range(orientations.shape[0]):
-            hsh_basis[z,i] = Z_numba( orientations[i,0], orientations[i,1], orientations[i,2], 
-                                     nlm[z,0], nlm[z,1], nlm[z,2] )
-    return hsh_basis
+        hsh_bf = np.empty((G_odf.shape[0]), np.complex128)
+        for i in range(G_odf.shape[0]):
+            hsh_bf[i] = Z_numba( G_odf[i,0], G_odf[i,1], G_odf[i,2], 
+                                     n, lm[z,0], lm[z,1] )        
+        for l in range(n_shsh):
+            shsh_basis_n[l] += np.real( symm_matrix_n[l, z] * hsh_bf )
+    return shsh_basis_n
+
+# @njit(parallel=True)
+# def Z_all_parallel(orientations, nlm):
+#     n_hsh = nlm.shape[0]
+#     hsh_basis = np.empty((n_hsh, orientations.shape[0]), np.complex128)
+#     for z in prange(n_hsh):
+#         for i in range(orientations.shape[0]):
+#             hsh_basis[z,i] = Z_numba( orientations[i,0], orientations[i,1], orientations[i,2], 
+#                                      nlm[z,0], nlm[z,1], nlm[z,2] )
+#     return hsh_basis
 
 @njit
 def Z_numba(omega, theta, phi, n, l, m):
@@ -495,25 +633,6 @@ def get_NsHSH(point_group, n):
                 NsHSH_all = eval( file.read() )[point_group]
         return NsHSH_all[int(n/2-1)]
 
-def get_symmetrization_matrix( orders, symmetry ):
-    # Number sHSHs for each n
-    n_sHSHs  =  np.array( [get_NsHSH(symmetry,n) for n in orders] ) 
-    slices_shsh = np.array(
-        [ [ n_sHSHs[:k].sum(), n_sHSHs[:k+1].sum() ] 
-            for k in range(n_sHSHs.shape[0]) ]) # 'slices' of coefficients as numbers for each n
-    n_HSHs = np.array([(n+1)**2 for n in orders])
-    slices_hsh = np.array(
-        [ [ n_HSHs[:k].sum(), n_HSHs[:k+1].sum() ] 
-            for k in range(n_HSHs.shape[0]) ]) # 'slices' of coefficients as numbers for each n
-    
-    # calculate the matrix that transforms HSHs to sHSHs
-    # Xsn = HSH_symmetrization_matrix(np.max(orders), symmetry)
-    Xsn = {'0': np.array([[1.]], np.complex128)}
-    for n in orders[1:]:
-        _, csym = cSymmHSH(symmetry, n) # get sSHSs and orders
-        Xsn[str(n)] = csym
-
-    return Xsn, slices_hsh, slices_shsh
 
 def get_orders( symmetry, n_max = 20, info=True, exclude_ghosts=True ):
     ''' Gives the allowed orders up to n_max
@@ -531,7 +650,7 @@ def get_orders( symmetry, n_max = 20, info=True, exclude_ghosts=True ):
         orders where sHSHs exist for this point group
     '''
     if info:
-        print('\t\tn\tNo of symmetrized HSHs for order n')
+        print('\t\torder\tNo of symmetrized HSHs')
     n_allowed = [0]
     for n in range(1,n_max+1):
         Nn = get_NsHSH(symmetry,n)
@@ -540,7 +659,7 @@ def get_orders( symmetry, n_max = 20, info=True, exclude_ghosts=True ):
             Nn = 0
         else:
             ghost = ''
-        if info:
+        if info and Nn > 0:
             print('\t\t%u\t %u%s' % (n, Nn, ghost) )
         if Nn > 0:
             n_allowed.append(n)
