@@ -1,6 +1,5 @@
 import numpy as np
-import scipy.sparse as sp
-from scipy.sparse import eye as speye
+# import scipy.sparse as sp
 from scipy.linalg import null_space, qr
 from numba import njit
 
@@ -16,13 +15,21 @@ from ..model import symmetries as sym
 #
 # Rewritten in python by Moritz Frewein 2025
 
+def check_matlab_matrix(M, tol):
+    data = np.genfromtxt('/Users/moritz/Documents/code/textom/textom/ressources/symmetrizedHSH/output/matrix.csv',
+                         delimiter=',',dtype=np.complex128)
+    print(np.allclose(data,M, atol=tol))
+
 def HSH_symmetrization_matrix(n_max, symmetry):
     gen = sym.generators(symmetry)
     gen_sample = sym.generators('1')
-    X = collect_symm(gen,gen_sample, n_max=n_max, precision=1e-12)
+    X = collect_symm(gen,gen_sample, n_max=n_max, 
+                    precision=1e-12
+                    # precision=1e-15
+                    )
     return X
 
-def collect_symm(gc, gs, n_max, precision):
+def collect_symm(gc, gs, n_max, precision, skip_ghosts=True):
     # collect_symm find sets of coefficients up to n_max that make the
     # hyperspherical harmonic expansion invariant to the cubic crystal point
     # group generators. These can  be interpreted as defining a compact basis
@@ -37,16 +44,24 @@ def collect_symm(gc, gs, n_max, precision):
     #   symm  - a cell array containing three columns which are the outputs X,
     #           L and M of the make_symm_pq function. The value of N starts at
     #           zero and increases with the row.
-    symm = np.empty((n_max + 1, 3), dtype=object)
 
-    n = 2
+    # symm = np.empty((n_max + 1, 3), dtype=object)
+    symm = {'0': np.array([[1.]], np.complex128)}
+
+    if skip_ghosts:
+        step=4
+    else:
+        step=2
+
+    n = step
     while n <= n_max:
         X, L, M = make_symm(gc, gs, n, precision)
         if X is not None and np.size(X) > 0:
-            symm[n, 0] = X
-            symm[n, 1] = L
-            symm[n, 2] = M
-        n += 2
+            # symm[n, 0] = X
+            # symm[n, 1] = L
+            # symm[n, 2] = M
+            symm[str(n)] = X.T
+        n += step
 
     return symm
 
@@ -108,7 +123,7 @@ def make_symm(gc, gs, n, TOL):
     # uncoupled to the coupled basis.
 
     dim = (n + 1) ** 2
-    CG = sp.lil_matrix((dim, dim))
+    CG = np.zeros((dim, dim))
 
     j_vals = np.arange(0, n + 1)
     for a in range(n + 1):
@@ -121,7 +136,7 @@ def make_symm(gc, gs, n, TOL):
                 col = (int(m1[c]) + l) * (2 * l + 1) + (int(m2[c]) + l) + 1
                 CG[row - 1, col - 1] = C[c]   # MATLAB → Python index shift
 
-    CG = CG.tocsr()
+    # CG = CG.tocsr()
     CR = complex_real(n)
 
     X = np.eye(dim) #speye(dim, format="csr")
@@ -165,37 +180,37 @@ def make_symm(gc, gs, n, TOL):
     Ul = U(l, la[0], la[1], la[2])
     Ur = U(l, -ra[0], ra[1], ra[2])
     Ra = np.round( np.kron(Ur, Ul), rou)
-    Y = sp_orth(null_space((Ra - np.eye(Ra.shape[0])).T,TOL))
-    Q = Y @ Y.T
-    X = sp_orth(null_space((P @ Q  - np.eye(Q.shape[0])).T,TOL))
-    P = X @ X.T
+    Y = orth(null_space((Ra - np.eye(Ra.shape[0])).T,TOL))
+    Q = Y @ Y.conj().T
+    X = orth(null_space(np.round((P @ Q  - np.eye(Q.shape[0])).T,rou),TOL))
+    P = X @ X.conj().T
 
     Ul = U(l, lb[0], lb[1], lb[2])
     Ur = U(l, -rb[0], rb[1], rb[2])
     Rb = np.round( np.kron(Ur, Ul), rou)
-    Y = sp_orth(null_space((Rb - np.eye(Rb.shape[0])).T,TOL))
-    Q = Y @ Y.T
-    X = sp_orth(null_space((P @ Q  - np.eye(Q.shape[0])).T,TOL))
-    P = X @ X.T
+    Y = orth(null_space((Rb - np.eye(Rb.shape[0])).T,TOL))
+    Q = Y @ Y.conj().T
+    X = orth(null_space(np.round((P @ Q  - np.eye(Q.shape[0])).T,rou),TOL))
+    P = X @ X.conj().T
 
     Ul = U(l, lc[0], lc[1], lc[2])
     Ur = U(l, -rc[0], rc[1], rc[2])
     Rc = np.round( np.kron(Ur, Ul), rou)
-    Y = sp_orth(null_space((Rc - np.eye(Rc.shape[0])).T,TOL))
-    Q = Y @ Y.T
-    X = sp_orth(null_space((P @ Q  - np.eye(Q.shape[0])).T,TOL))
-    P = X @ X.T
+    Y = orth(null_space((Rc - np.eye(Rc.shape[0])).T,TOL))
+    Q = Y @ Y.conj().T
+    X = orth(null_space(np.round((P @ Q  - np.eye(Q.shape[0])).T,rou),TOL))
+    P = X @ X.conj().T
 
     Ul = U(l, ld[0], ld[1], ld[2])
     Ur = U(l, -rd[0], rd[1], rd[2])
     Rd = np.round( np.kron(Ur, Ul), rou)
-    Y = sp_orth(null_space((Rd - np.eye(Rd.shape[0])).T,TOL))
-    Q = Y @ Y.T
-    X = sp_orth(null_space((P @ Q  - np.eye(Q.shape[0])).T,TOL))
+    Y = orth(null_space((Rd - np.eye(Rd.shape[0])).T,TOL))
+    Q = Y @ Y.conj().T
+    X = null_space(np.round((P @ Q  - np.eye(Q.shape[0])).T,rou),TOL)
 
     # convert to coupled basis
     X = CG @ X
-    X = np.array(CR @ clean(CR.T @ X, TOL))
+    X = np.array(CR @ clean(CR.conj().T @ np.round(X,rou), TOL))
 
     # numerical error checks
     def max_err(R):
@@ -292,6 +307,22 @@ def U(j,ome,tta,phi,invert=False):
     for a in range(n):
         U[a, :] = U[a, :] * r_base**m[a]
         U[:, a] = U[:, a] * c_base**m[a]
+
+    # tmp = tan(w / 2.) * cos(th);
+    # tmp = (1. - 1i * tmp) / realsqrt(1 + tmp^2.);
+    # r_base = 1i * exp(-1i * ph) * tmp;
+    # c_base = -1i * exp(1i * ph) * tmp;
+    
+    # % Require w to be in [-\pi, \pi]
+    # w = mod(w + pi, 2. * pi) - pi;
+    # xi = 2. * asin(sin(w / 2.) * sin(th));
+    # U = wigner_little_d(j, xi);
+    
+    # m = -j:j;
+    # n = 2 * j + 1;
+    # for a = 1:n
+    #     U(a, :) = U(a, :) * r_base^m(a);
+    #     U(:, a) = U(:, a) * c_base^m(a);
     return U
 
 @njit
@@ -376,147 +407,145 @@ def rotation_mat(j, w, th, ph):
 #     N[np.abs(N) < TOL] = 0.0
 #     return N
 
-# import numpy as np
-# import scipy.sparse as sp
-# import scipy.sparse.linalg as spla
-from scipy.linalg import lu
+# # import numpy as np
+# # import scipy.sparse as sp
+# # import scipy.sparse.linalg as spla
+# from scipy.linalg import lu
+
+# def sp_null(A, opt, TOL=1e-12):
+#     """
+#     Find left or right null space of a sparse matrix A.
+
+#     Parameters
+#     ----------
+#     A : (m, n) sparse or dense array
+#         Input matrix.
+#     opt : int
+#         0 -> left null space, 1 -> right null space.
+#     TOL : float
+#         Threshold for considering pivots as significant.
+
+#     Returns
+#     -------
+#     N : ndarray
+#         Left or right null space basis (not orthonormal).
+#     """
+#     if not sp.isspmatrix(A):
+#         A = sp.csr_matrix(A)
+
+#     if opt == 0:
+#         # Zero-out tiny entries
+#         A = A.tocoo()
+#         mask = np.abs(A.data) >= TOL
+#         A = sp.coo_matrix((A.data[mask], (A.row[mask], A.col[mask])), shape=A.shape).tocsr()
+
+#         L, U, Q = luq(A, TOL)
+
+#         diagU = np.abs(U.diagonal())
+#         r = np.count_nonzero(diagU > TOL)
+
+#         inv_L = np.linalg.inv(L.toarray())  # L is sparse but invert to dense
+#         N = inv_L[r:, :]
+
+#         # clean small values
+#         N[np.abs(N) < TOL] = 0.0
+#         return N
+
+#     elif opt == 1:
+#         return sp_null(A.T, 0, TOL).T
+#     else:
+#         raise ValueError("Invalid option. Use 0 (left) or 1 (right).")
 
 
-def sp_null(A, opt, TOL=1e-12):
-    """
-    Find left or right null space of a sparse matrix A.
+# def luq(A, TOL=1e-12):
+#     """
+#     Perform LU-like factorization with permutation to construct
+#     A = L * U * Q with block-triangular U.
 
-    Parameters
-    ----------
-    A : (m, n) sparse or dense array
-        Input matrix.
-    opt : int
-        0 -> left null space, 1 -> right null space.
-    TOL : float
-        Threshold for considering pivots as significant.
+#     Returns
+#     -------
+#     L : sparse matrix (m x m)
+#     U : sparse matrix (m x n)
+#     Q : sparse matrix (n x n)
+#     """
+#     if not sp.isspmatrix(A):
+#         A = sp.csr_matrix(A)
 
-    Returns
-    -------
-    N : ndarray
-        Left or right null space basis (not orthonormal).
-    """
-    if not sp.isspmatrix(A):
-        A = sp.csr_matrix(A)
+#     m, n = A.shape
 
-    if opt == 0:
-        # Zero-out tiny entries
-        A = A.tocoo()
-        mask = np.abs(A.data) >= TOL
-        A = sp.coo_matrix((A.data[mask], (A.row[mask], A.col[mask])), shape=A.shape).tocsr()
+#     # Dense LU decomposition (SciPy's sparse lu does not expose P, Q in same way)
+#     A_dense = A.toarray()
+#     P, L, U = lu(A_dense)  # A = P @ L @ U
 
-        L, U, Q = luq(A, TOL)
+#     # Convert to A = L * U * Q form
+#     if n < m:
+#         p = m - n
+#         L = np.hstack([L, np.vstack([np.zeros((n, p)), np.eye(p)])])
+#         U = np.vstack([U, np.zeros((p, n))])
 
-        diagU = np.abs(U.diagonal())
-        r = np.count_nonzero(diagU > TOL)
+#     L = P.T @ L
+#     Q = np.eye(n)
 
-        inv_L = np.linalg.inv(L.toarray())  # L is sparse but invert to dense
-        N = inv_L[r:, :]
+#     # Identify pivots
+#     diagU = np.abs(np.diag(U))
+#     p1 = np.where(diagU > TOL)[0]
+#     p = len(p1)
 
-        # clean small values
-        N[np.abs(N) < TOL] = 0.0
-        return N
+#     # Permute rows and columns
+#     r2 = np.setdiff1d(np.arange(m), p1)
+#     r_perm = np.concatenate([p1, r2])
+#     L = L[:, r_perm]
+#     U = U[r_perm, :]
 
-    elif opt == 1:
-        return sp_null(A.T, 0, TOL).T
-    else:
-        raise ValueError("Invalid option. Use 0 (left) or 1 (right).")
+#     c2 = np.setdiff1d(np.arange(n), p1)
+#     c_perm = np.concatenate([p1, c2])
+#     U = U[:, c_perm]
+#     Q = Q[c_perm, :]
 
+#     # Update indices
+#     p1 = np.arange(p)
+#     r2 = np.arange(p, m)
+#     c2 = np.arange(p, n)
 
-def luq(A, TOL=1e-12):
-    """
-    Perform LU-like factorization with permutation to construct
-    A = L * U * Q with block-triangular U.
+#     # Eliminate U21 block
+#     if p > 0 and r2.size > 0:
+#         X = np.linalg.solve(U[p1, p1], U[p1, :][:, c2])
+#         U[np.ix_(p1, c2)] = 0.0
+#         Q[p1, :] = Q[p1, :] + X @ Q[c2, :]
 
-    Returns
-    -------
-    L : sparse matrix (m x m)
-    U : sparse matrix (m x n)
-    Q : sparse matrix (n x n)
-    """
-    if not sp.isspmatrix(A):
-        A = sp.csr_matrix(A)
+#         X = np.linalg.solve(U[p1, p1], U[np.ix_(r2, p1)].T).T
+#         L[:, p1] = L[:, p1] + L[:, r2] @ X
+#         U[np.ix_(r2, p1)] = 0.0
+#         U[np.ix_(r2, c2)] -= X @ U[np.ix_(p1, c2)]
 
-    m, n = A.shape
+#     # Clean up block structure
+#     if r2.size > 0 and c2.size > 0:
+#         r3 = p + np.where(np.max(np.abs(U[np.ix_(r2, c2)]), axis=1) > TOL)[0]
+#         c3 = p + np.where(np.max(np.abs(U[np.ix_(r2, c2)]), axis=0) > TOL)[0]
 
-    # Dense LU decomposition (SciPy's sparse lu does not expose P, Q in same way)
-    A_dense = A.toarray()
-    P, L, U = lu(A_dense)  # A = P @ L @ U
+#         if r3.size > 0 or c3.size > 0:
+#             # Recursively decompose U22
+#             Lr, Ur, Qr = luq(sp.csr_matrix(U[np.ix_(r3, c3)]), TOL)
+#             U[np.ix_(r3, c3)] = Ur
+#             L[:, r3] = L[:, r3] @ Lr
+#             Q[c3, :] = Qr @ Q[c3, :]
 
-    # Convert to A = L * U * Q form
-    if n < m:
-        p = m - n
-        L = np.hstack([L, np.vstack([np.zeros((n, p)), np.eye(p)])])
-        U = np.vstack([U, np.zeros((p, n))])
+#             q = np.count_nonzero(np.abs(np.diag(U)) > TOL)
+#             U = np.block([
+#                 [U[:q, :q], np.zeros((q, n - q))],
+#                 [np.zeros((m - q, q)), np.zeros((m - q, n - q))]
+#             ])
+#         else:
+#             U[np.ix_(r2, c2)] = 0.0
 
-    L = P.T @ L
-    Q = np.eye(n)
+#     return sp.csr_matrix(L), sp.csr_matrix(U), sp.csr_matrix(Q)
 
-    # Identify pivots
-    diagU = np.abs(np.diag(U))
-    p1 = np.where(diagU > TOL)[0]
-    p = len(p1)
-
-    # Permute rows and columns
-    r2 = np.setdiff1d(np.arange(m), p1)
-    r_perm = np.concatenate([p1, r2])
-    L = L[:, r_perm]
-    U = U[r_perm, :]
-
-    c2 = np.setdiff1d(np.arange(n), p1)
-    c_perm = np.concatenate([p1, c2])
-    U = U[:, c_perm]
-    Q = Q[c_perm, :]
-
-    # Update indices
-    p1 = np.arange(p)
-    r2 = np.arange(p, m)
-    c2 = np.arange(p, n)
-
-    # Eliminate U21 block
-    if p > 0 and r2.size > 0:
-        X = np.linalg.solve(U[p1, p1], U[p1, :][:, c2])
-        U[np.ix_(p1, c2)] = 0.0
-        Q[p1, :] = Q[p1, :] + X @ Q[c2, :]
-
-        X = np.linalg.solve(U[p1, p1], U[np.ix_(r2, p1)].T).T
-        L[:, p1] = L[:, p1] + L[:, r2] @ X
-        U[np.ix_(r2, p1)] = 0.0
-        U[np.ix_(r2, c2)] -= X @ U[np.ix_(p1, c2)]
-
-    # Clean up block structure
-    if r2.size > 0 and c2.size > 0:
-        r3 = p + np.where(np.max(np.abs(U[np.ix_(r2, c2)]), axis=1) > TOL)[0]
-        c3 = p + np.where(np.max(np.abs(U[np.ix_(r2, c2)]), axis=0) > TOL)[0]
-
-        if r3.size > 0 or c3.size > 0:
-            # Recursively decompose U22
-            Lr, Ur, Qr = luq(sp.csr_matrix(U[np.ix_(r3, c3)]), TOL)
-            U[np.ix_(r3, c3)] = Ur
-            L[:, r3] = L[:, r3] @ Lr
-            Q[c3, :] = Qr @ Q[c3, :]
-
-            q = np.count_nonzero(np.abs(np.diag(U)) > TOL)
-            U = np.block([
-                [U[:q, :q], np.zeros((q, n - q))],
-                [np.zeros((m - q, q)), np.zeros((m - q, n - q))]
-            ])
-        else:
-            U[np.ix_(r2, c2)] = 0.0
-
-    return sp.csr_matrix(L), sp.csr_matrix(U), sp.csr_matrix(Q)
-
-
-
-def sp_orth(A):
+def orth(A):
     """
     Compute an orthonormal basis for the columns of A.
     """
-    Q,_ = qr(A, mode='economic')  # 'economic' = reduced QR
+    Q,_,_ = qr(A, mode='economic', pivoting=True)  # 'economic' = reduced QR
+    # Q,_ = qr(A, mode='economic')  # 'economic' = reduced QR
     return Q
 
 def complex_real(n):
@@ -533,7 +562,8 @@ def complex_real(n):
 
     sqrt2 = np.sqrt(2.0)
     dim = (n + 1) ** 2
-    U = sp.lil_matrix((dim, dim), dtype=complex)
+    # U = sp.lil_matrix((dim, dim), dtype=complex)
+    U = np.zeros((dim, dim), np.complex128)
 
     for L in range(n + 1):
         row0 = (L + 1)**2 - L - 1  # MATLAB 1-based → Python 0-based
@@ -543,15 +573,16 @@ def complex_real(n):
         for M in range(1, L + 1):
             row_minus = row0 - M
             row_plus  = row0 + M
-            col1 = col0 + 2*M - 2
-            col2 = col0 + 2*M - 1
+            col1 = col0 + 2*M - 1
+            col2 = col0 + 2*M
 
             U[row_minus, col1] = (-1)**M * (1j)**L / sqrt2
             U[row_plus,  col1] = (1j)**L / sqrt2
             U[row_minus, col2] = (-1)**(M-1) * (1j)**(L-1) / sqrt2
             U[row_plus,  col2] = (1j)**(L-1) / sqrt2
 
-    return U.tocsr()
+    return U
+    # return U.tocsr()
 
 def clean_col(A, a, TOL):
     """
@@ -590,30 +621,36 @@ def clean(A, TOL=1e-12):
     %   A   - matrix whose columns form an orthonormal column space for A.
     """
     m, n = A.shape
+    rou = int(-np.log10(TOL))
 
     # Initial processing
     for a in range(n):
         norm = np.linalg.norm(A[:, a])
         if norm > 0:
             A[:, a] /= norm
-        A = clean_col(A, a, TOL)
+        # A = clean_col(A, a, TOL)
 
     # Forward sweep
     r1 = 0
     for a in range(n):
+        # Find column containing most significant entry
         p = 0.0
         while p < TOL and r1 < m:
             r1 += 1
             slice_abs = np.abs(A[r1-1, a:n])
             p_idx = np.argmax(slice_abs)
             p = slice_abs[p_idx]
-        b = (a) + p_idx
+        b = a + p_idx
 
+        # Should not happen
         if r1 >= m:
-            # print("A is not full column rank")
+            print("A is not full column rank")
             break
 
-        # Swap column a with column b
+        # Swap column into leading position (Swap column a with column b)
+        # swap = A[:,a].copy()
+        # A[:,a] = A[:,b]
+        # A[:,b] = swap
         A[:, [a, b]] = A[:, [b, a]]
 
         # Make leading entry real
@@ -621,24 +658,30 @@ def clean(A, TOL=1e-12):
         if np.abs(lead) > 0:
             A[:, a] *= np.conj(lead) / np.abs(lead)
 
-        # Cancel leading entries in remaining columns
-        for col_idx in range(a+1, n):
-            if np.abs(A[r1-1, col_idx]) > TOL:
-                A[:, col_idx] -= (A[r1-1, col_idx] / A[r1-1, a]) * A[:, a]
-                A = clean_col(A, col_idx, TOL)
+        # Cancel leading entries, breaks orthogonality and normality
+        cols = np.where(np.abs(A[r1-1, a+1:n]) > TOL)[0] + (a+1)
+        for col_idx in cols:
+            factor = A[r1-1, col_idx] / A[r1-1, a]
+            A[:, col_idx] -= factor * A[:, a]
+            # A = clean_col(A, col_idx, TOL)
+        # for col_idx in range(a+1, n):
+        #     if np.abs(A[r1-1, col_idx]) > TOL:
+        #         A[:, col_idx] -= (A[r1-1, col_idx] / lead) * A[:, a]
+                # A = clean_col(A, col_idx, TOL)
 
     # Backward sweep to restore orthonormality
+    # Q, _, _ = qr(np.fliplr(A), mode='economic', pivoting=True)
     Q, _ = qr(np.fliplr(A), mode='economic')
-    A = np.fliplr(Q)
+    A = np.round(np.fliplr(Q), rou) # this is my "clean_col"
 
     for a in range(n):
-        A = clean_col(A, a, TOL)
+        # A = clean_col(A, a, TOL)
         # Make leading entry positive
         nz = np.flatnonzero(A[:, a])
         if len(nz) > 0:
             A[:, a] *= np.sign(A[nz[0], a])
 
-    return A
+    return np.round(A, rou)
 
 @njit
 def CleGor(j1, j2, J, M):
