@@ -1,12 +1,15 @@
-import os
+import os, re
 import numpy as np
 from time import time
 import h5py
 
 # domestic
-from ..misc import sparsemult, integrate_c
-from ...config import data_type
+from ..numba_plugins import sparsemult, integrate_c, masked_mean_axis0
+from ..misc import check_version
+from ...config import data_type, hex_notation
 from ..model.model_textom import model_textom
+from ..model.symmetries import miller_to_hex
+from ..data_treatment.data import load_peak_regions
 from ..odf import hsh
 from ..exceptions import dimension_error
 
@@ -21,7 +24,7 @@ class fitting:
 
     Takes input from a model and a data class object or a previously saved 'fit' pickle
     """
-    def __init__(self, path, mod:model_textom, rec1d=False ):
+    def __init__(self, sample_dir, mod:model_textom, rec1d=False, sf_from_data=False, set_negative_zero=False ):
         """
 
         Parameters
@@ -35,22 +38,37 @@ class fitting:
         ------------
 
         """
-        print('Initialize Fit')
+        print('Initializing Fit')
         datafile = 'data_1drec.h5' if rec1d else 'data_textom.h5'
         # import stuff from data and mod objects
         t0=time()
-        with h5py.File( os.path.join(path,'analysis',datafile), 'r') as hf:
-            self.data = hf['data'][()].astype(data_type)
+        with h5py.File( os.path.join(sample_dir,'analysis',datafile), 'r') as hf:
+            self.data_unmasked = hf['data'][()].astype(data_type)
             self.scanmask = hf['scanmask'][()]
             self.gt_mask = hf['gt_mask'][()]
             if not rec1d:
-                self.peak_reg = hf['peak_reg'][()]
-                self.q = hf['q'][()]
-                self.hkl = [h.decode('utf-8') for h in hf['hkl']]
-            # self.mask_detector = hf['mask_detector'][()]
+                self.q_full = hf['q'][()]
+                self.mask_detector = hf['mask_peaks'][()]
+                
+                # self.peak_reg = hf['peak_reg'][()]
+                # self.hkl = [h.decode('utf-8') for h in hf['hkl']]
+                # i'll take these 2 out of the txt file then it can be modified afterwards
+                # and we load the peaks that are written inside
+                path_peak_regions = os.path.join(sample_dir,'analysis','peak_regions.txt')
+                self.peak_reg, self.hkl_full = load_peak_regions(path_peak_regions, exclude_empty=True)
+
+            # check_version(hf['textom_version'][()])
+        if set_negative_zero:
+            self.data_unmasked[self.data_unmasked<0] = 0
         print('\tLoaded Data')
 
-        self.mask_detector = np.genfromtxt(os.path.join(path,'analysis','fit_detmask.txt'),bool)
+        self.mod = mod
+        self.flag_use_peak = np.ones(len(self.hkl_full), bool)
+        self.mask_data()
+        self.list_peaks()
+        self.sf_from_data = sf_from_data
+
+        # self.mask_detector = np.genfromtxt(os.path.join(path,'analysis','fit_detmask.txt'),bool)
         
         # ######### do check if the test for changing detector angle direction
         # nD = self.data.shape[0]
@@ -68,13 +86,6 @@ class fitting:
         tomogram = mod.tomogram.flatten()/mod.tomogram.max() # normalize so c0 will be around 1
         tomogram[tomogram<0] = 0
 
-        self.classic = mod.classic
-        if mod.classic:
-            self.C = np.zeros((tomogram.size,1), data_type)
-            self.C[mod.mask_voxels, 0] = tomogram[mod.mask_voxels] # initial guess for c0
-        else:
-            self.C = np.zeros((tomogram.size,mod.difflets.shape[1]),data_type)
-        self.ns=[0]
         self.title, self.Beams, self.iBeams = mod.title, mod.Beams, mod.iBeams
 
         if not rec1d:
@@ -86,22 +97,44 @@ class fitting:
                 mod.iBeams[g, t, :iend], tomogram )
             # the simulated intensity will be multiplied by this to have C~O(1)
             gt_dat = np.argmin( np.linalg.norm(self.gt_mask - np.array([g,t]), axis=1) )
-            self.sim_fac = self.data[gt_dat].max()/mod.difflets[0].max() / T
-            if mod.classic:
-                # self.difflet0_1D = mod.difflets[0].reshape(mod.detShape)[:,0] # not used
-                self.set_orders(mod, 0, info=False)
+            if sf_from_data:
+                self.sim_fac = 1 / T
             else:
-                self.update_mod(mod)
+                self.sim_fac = self.data[gt_dat].max()/mod.difflets[0].max() / T
+            self.dat_scal = self.data[gt_dat].max()/T # and then do the rest in update_mod
+            self.update_mod()
 
             # self.q = mod.Qq_det.reshape(mod.detShape)[:,0] # 1D q-vector
-            self.q = np.array([qr.mean() for qr in self.peak_reg])
+            # self.q = np.array([qr.mean() for qr in self.peak_reg])
             try:
                 self.chi = mod.Chi_det.reshape(mod.detShape)[:,0]
             except:
                 self.chi = mod.chi
 
+        self.C = np.zeros((self.mod.x_p.shape[0], self.mod.difflets.shape[1]), data_type)
         self.choose_projections(info=False)
 
+    def mask_data( self ):
+        self.q = self.q_full[self.flag_use_peak]
+        self.data = self.data_unmasked[:,:,self.flag_use_peak][:,self.mask_detector[:,self.flag_use_peak]]
+
+    def toggle_peak_use( self, peak_idcs=[], flags=None ):
+        if flags is not None:
+            self.flag_use_peak = flags
+        else:
+            peak_idcs = np.atleast_1d(peak_idcs).astype(int)
+            for p in peak_idcs:
+                self.flag_use_peak[p] = not self.flag_use_peak[p]
+        self.mask_data()
+        self.update_mod()
+        self.list_peaks()
+
+    def list_peaks( self ):
+        print('\tAvailable peaks:')
+        print('\tIdx\tfit_flag\tq-range\t\thkl')
+        for p in range(len(self.hkl_full)):
+            print(f'\t{p}\t{self.flag_use_peak[p]}\t\t{self.peak_reg[p,0]:.2f}\t{self.peak_reg[p,1]:.2f}\t{self.hkl_full[p]}')
+
     def prepare_fit_c0( self, peak=None ):
         """ Calculates 0-dimensional data for fast estimation of the zero order 
         
@@ -112,9 +145,10 @@ class fitting:
             isotropic as possible), if None uses the whole dataset, default None
         """
         if peak:
+            # not sure if this is still functional
             peak_start_idx = self.mask_detector[:peak*self.detShape[1]].sum()
             peak_end_idx = self.mask_detector[:(peak+1)*self.detShape[1]].sum()
-            self.total_im = self.data[:,peak_start_idx:peak_end_idx].sum(axis=1) # total peak intensity for each image
+            self.total_im = self.data[:,:,peak].sum(axis=1).sum(axis=1) # total peak intensity for each image
             self.data_0D = self.total_im / self.difflets_rot_masked[0,0,peak_start_idx:peak_end_idx].sum() 
         else:
             self.total_im = self.data.sum(axis=1) # total intensity for each image
@@ -175,7 +209,7 @@ class fitting:
         else:
             self.C = c_tmp[:,:self.C.shape[1]]
 
-    def update_mod(self, mod:model_textom):
+    def update_mod(self):
         """ Imports stuff from the model class
 
         Parameters
@@ -183,57 +217,54 @@ class fitting:
         mod : model class object
         """
         # regroup diffractlets into peaks and mask them
-        self.difflets_rot_masked = np.empty( (*mod.difflets.shape[:2], self.mask_detector.sum()), data_type )
-        self.detShape = (mod.detShape[0], self.peak_reg.shape[0]) # effective detector shape after choosing peak regions
-        self.odf_mode = mod.odf_mode
-        self.odf_par = mod.odf_parameter
+        n_peaks = self.flag_use_peak.sum()
+        self.hkl = np.array(self.hkl_full)[self.flag_use_peak]
+        self.difflets_rot_masked = np.empty( (*self.mod.difflets.shape[:2], self.mask_detector[:,self.flag_use_peak].sum()), data_type )
+        self.detShape = (self.mod.detShape[0], n_peaks) # effective detector shape after choosing peak regions
+        self.odf_mode = self.mod.odf_mode
+        self.odf_par = self.mod.odf_parameter
         if self.odf_mode == 'hsh':
-            self.Sn = hsh.get_order_slices( self.odf_par, mod.symmetry )
+            self.Sn = hsh.get_order_slices( self.odf_par, self.mod.symmetry )
 
-        q_det = mod.Qq_det.reshape(mod.detShape)[0]
-        for g, difflets_g in enumerate(mod.difflets):
+        if self.sf_from_data:
+            # take structure factor out of data:
+            data_chimean = np.mean([masked_mean_axis0(d, self.mask_detector) for d in self.data_unmasked.reshape((self.data.shape[0],*self.detShape))], axis=0)
+
+        for g, difflets_g in enumerate(self.mod.difflets):
             for l, dlet in enumerate(difflets_g):
                 dlet_tmp = np.empty( self.detShape, data_type)
-                for k, reg in enumerate(self.peak_reg):
-                    mask_reg = ((q_det >= reg[0]) & (q_det <= reg[1]))
-                    dlet_tmp[:,k] = dlet[:,mask_reg].sum(axis=1)
-                try:
-                    self.difflets_rot_masked[g,l] = dlet_tmp.flatten()[self.mask_detector]
-                except:
-                    raise dimension_error('Check if No of peak regions x No of azimuthal bins agree with the detector mask size')
-        self.difflets_rot_masked *= self.sim_fac#(_rotate_difflets_stack(mod.Rs, difflets) * self.sim_fac).astype(data_type)
+                for k in range(n_peaks):
+                    # if hex_notation and self.mod.symmetry[0] == '6':
+                    #     hkls = np.array([miller_to_hex(hkl) for hkl in self.mod.hkl])
+                    # else:
+                    #     hkls = self.mod.hkl
+                    hkls = np.column_stack([ # this should be safe for miller and hex notation
+                        self.mod.hkl[:,0],self.mod.hkl[:,1],self.mod.hkl[:,-1]
+                        ])
 
-        self.C = np.zeros((mod.x_p.shape[0], mod.difflets.shape[1]), data_type)
-        # print('\tupdated fit object')
+                    # check which peaks are specified in the file for this data bin
+                    coords = re.findall(r'\[([-?\d, ]+)\]', self.hkl[k])
+                    matches = []
+                    for c in coords:
+                        vals = np.fromstring(c, sep=',', dtype=int)
+                        vals = np.array([vals[0], vals[1], vals[-1]])
+                        # find where in myarray this triplet appears
+                        idx = np.where((hkls == vals).all(axis=1))[0]
+                        # matches.extend(idx.tolist())
+                        matches.append(idx)
+                    mask_peaks = np.zeros(self.mod.hkl.shape[0], bool)
+                    mask_peaks[matches] = True
 
-    def update_mod_old(self, mod:model_textom):
-        """ Imports stuff from the model class
+                    dlet_tmp[:,k] = dlet[:,mask_peaks].sum(axis=1) # sum over all peaks that are in the bin
 
-        Parameters
-        ----------
-        mod : model class object
-        """
-        self.Sn = mod.Sn
-        self.ns = mod.ns
-        # update diffractlets:
-        difflets_full = mod.difflets.reshape((mod.difflets.shape[0],*mod.detShape))
-        # regroup diffractlets into peaks, rotate and mask them
-        difflets = np.empty( (mod.difflets.shape[0], self.mask_detector.sum()), data_type )
-        self.detShape = (self.peak_reg.shape[0], mod.detShape[1])
-        q_det = mod.Qq_det.reshape(mod.detShape)[:,0]
-        for l, dlet in enumerate(difflets_full):
-            dlet_tmp = np.empty( self.detShape, data_type)
-            for k, reg in enumerate(self.peak_reg):
-                mask_reg = ((q_det >= reg[0]) & (q_det <= reg[1]))
-                dlet_tmp[k] = dlet[mask_reg].sum(axis=0)
-            try:
-                difflets[l] = dlet_tmp.flatten()[self.mask_detector]
-            except:
-                raise dimension_error('Check if No of peak regions x No of azimuthal bins agree with the detector mask size')
-        self.difflets_rot_masked = (_rotate_difflets_stack(mod.Rs, difflets) * self.sim_fac).astype(data_type)
-
-        self.C = np.zeros((mod.x_p.shape[0], mod.difflets.shape[0]), data_type)
-        # print('\tupdated fit object')
+                    if self.sf_from_data:
+                        dlet_tmp[:,k] *= data_chimean[k] / self.mod.structure_factors[mask_peaks].sum()
+                # try:
+                self.difflets_rot_masked[g,l] = dlet_tmp[self.mask_detector[:,self.flag_use_peak]]
+                    # self.difflets_rot_masked[g,l] = dlet_tmp[self.mask_detector]
+                # except:
+                #     raise dimension_error('Check if No of peak regions x No of azimuthal bins agree with the detector mask size')
+        self.difflets_rot_masked *= self.dat_scal / self.difflets_rot_masked[0].max()
 
     def choose_projections(self, mode='full', info=True):
         """ Decides which sample rotations to use for fitting
@@ -457,6 +488,7 @@ class fitting:
         c_proj = integrate_c( self.Beams[g,t,:iend], self.iBeams[g,t,:iend], C )
         # calculate the image
         image = c_proj @ self.difflets_rot_masked[g] 
+        # image = c_proj @ self.difflets_rot[g] 
         return image
 
     def MDL( self, loss ):
@@ -481,7 +513,7 @@ class fitting:
         #     [beam[0]<2**23-1 for beam_g in self.iBeams[self.chosen_proj] for beam in beam_g])
         # im_mask = np.logical_and( lowmask, self.scanmask[self.chosen_proj].flatten() )
         N_images = np.sum( self.scanmask )
-        N_data = N_images * self.mask_detector.sum()
+        N_data = N_images * self.mask_detector[:,self.flag_use_peak].sum()
         # the number of parameters
         n_par = self.C[self.C[:,0]>0].size
         # calculate the medium description length
@@ -493,21 +525,31 @@ class fitting:
         d_out[self.scanmask] = data
         return d_out
     
+    def insert_fov_1d( self, data ):
+        d_out = np.zeros( (*self.scanmask.shape, data.shape[1]), data_type )
+        d_out[self.scanmask] = data
+        return d_out
+    
+    def insert_images( self, images_masked ):
+        images = np.zeros((images_masked.shape[0],*self.detShape), images_masked.dtype)
+        images[:,self.mask_detector[:,self.flag_use_peak]] = images_masked
+        return images
+    
     def check_orientations( self, g ):
 
-        proj = np.zeros( (self.Beams.shape[1], self.mask_detector.size), data_type )
-        proj[:,self.mask_detector] = self.projection( g, self.C, info=False )
+        proj = np.zeros( (self.Beams.shape[1], *self.mask_detector[:,self.flag_use_peak].shape), data_type )
+        proj[:,self.mask_detector[:,self.flag_use_peak]] = self.projection( g, self.C, info=False )
         dat = np.zeros_like(proj)
         d_mskd = self.data[ np.where(self.gt_mask[:,0]==g)[0] ]
         for nt, t in enumerate(self.gt_mask[self.gt_mask[:,0]==g][:,1]):
-            dat[t,self.mask_detector] = d_mskd[nt]
+            dat[t,self.mask_detector[:,self.flag_use_peak]] = d_mskd[nt]
 
         ori_dat = np.empty( (self.q.size, self.Beams.shape[1]), data_type )
         ori_fit = ori_dat.copy()
         for i_peak, q in enumerate(self.q):
             chi_hlf = self.chi[:int(self.chi.size/2)]
-            ori_dat[i_peak] = _get_mean_ori( dat, self.scanmask[g], i_peak, chi_hlf, self.detShape )
-            ori_fit[i_peak] = _get_mean_ori( proj, self.scanmask[g], i_peak, chi_hlf, self.detShape )
+            ori_dat[i_peak] = _get_mean_ori( dat, self.scanmask[g], i_peak, chi_hlf )
+            ori_fit[i_peak] = _get_mean_ori( proj, self.scanmask[g], i_peak, chi_hlf )
 
         ori_dat[ori_dat==0]=np.nan
         ori_fit[ori_fit==0]=np.nan
@@ -883,11 +925,11 @@ def _rotate_difflets_stack( Rs_stack, difflets ):
     return difflets_stack
 
 @njit(parallel=True)
-def _get_mean_ori( im, scanmask, i_peak, chi, detShape ):
+def _get_mean_ori( im, scanmask, i_peak, chi ):
     ori = np.zeros( im.shape[0], data_type )
     for t in prange(im.shape[0]):
         if scanmask[t]:
-            im2D = im[t].reshape( detShape )
+            im2D = im[t]#.reshape( detShape )
             # get azimuthal data for one peak
             I_azi = im2D[:,i_peak]
             nphi_2 = int( I_azi.size/2 ) 
