@@ -12,7 +12,7 @@ from ...version import __version__
 
 def optimize( fit:fitting, mode, alg='simple', 
             project_c0=True, project_all=False,
-            step_size_0=1e-5, 
+            step_size_0=1e-5, optimize_stepsize=True,
             tol=1e-3, minstep=1e-2, itermax=3000,
             name='', ):
     """ Gradient descent optimizer for texture tomography
@@ -101,6 +101,7 @@ def optimize( fit:fitting, mode, alg='simple',
                 return fit.grad_full( C )
             def jac_calib( C ):
                 return fit.grad_allbut0( C )
+            # do i need jac_calib? don't i just need to take grad[:,1:] ??
             
     # Initialize functions    
     loss = fun( c )
@@ -115,47 +116,53 @@ def optimize( fit:fitting, mode, alg='simple',
     invL[L!=0] = 2/L[L!=0]
     gam = np.tile( invL, (1,c.shape[1]) )
     direction = - gam * grad
-    step_size, _ = double_line_search(fit.loss,fit.C,direction,step_size_0,
-                            max_steps=100,tol=1e-6)
-
     t2 = time()
-    if step_size == 0:
-        print('\tOptimial stepsize not found, lower step_size_0 by a few orders of magnitude')
-        return 0
-    else:
-        print(f'\tOptimized stepsize: {step_size:.3e}, {t2-t1:.3f} s' )
-
-    if mode==4:
-        grad_2 = jac_calib( c )
-       
-        # calculate inverse lipschitz constants
-        L = fit.lipschitz().reshape((c.shape[0],1))
-        L[ L < 1e-4*L.max() ] = 0.
-        invL = np.zeros_like(L)
-        invL[L!=0] = 2/L[L!=0]
-        gam_2 = np.tile( invL, (1,c.shape[1]) )
-        direction_2 = - gam_2 * grad_2
-        step_size_2, _ = line_search(fit.loss,fit.C,direction_2,step_size,
+    if optimize_stepsize:
+        step_size, _ = double_line_search(fit.loss,fit.C,direction,step_size_0,
                                 max_steps=100,tol=1e-6)
-        t3 = time()
+
+        t2 = time()
         if step_size == 0:
             print('\tOptimial stepsize not found, lower step_size_0 by a few orders of magnitude')
             return 0
         else:
-            print(f'\tOptimized higher order stepsize: {step_size_2:.3e}, {t3-t2:.3f} s' )
+            print(f'\tOptimized stepsize: {step_size:.3e}, {t2-t1:.3f} s' )
 
-        step_size = np.column_stack([
-            step_size * np.ones_like(fit.C[:,0]),
-            step_size_2 * np.ones_like(fit.C[:,1:])
-        ])
-        t2=t3
+        if mode==4:
+            grad_2 = jac_calib( c )
+        
+            # calculate inverse lipschitz constants
+            L = fit.lipschitz().reshape((c.shape[0],1))
+            L[ L < 1e-4*L.max() ] = 0.
+            invL = np.zeros_like(L)
+            invL[L!=0] = 2/L[L!=0]
+            gam_2 = np.tile( invL, (1,c.shape[1]) )
+            direction_2 = - gam_2 * grad_2
+            step_size_2, _ = line_search(fit.loss,fit.C,direction_2,step_size,
+                                    max_steps=100,tol=1e-6)
+            t3 = time()
+            if step_size == 0:
+                print('\tOptimial stepsize not found, lower step_size_0 by a few orders of magnitude')
+                return 0
+            else:
+                print(f'\tOptimized higher order stepsize: {step_size_2:.3e}, {t3-t2:.3f} s' )
+
+            step_size = np.column_stack([
+                step_size * np.ones_like(fit.C[:,0]),
+                step_size_2 * np.ones_like(fit.C[:,1:])
+            ])
+            t2=t3
+    else:
+        step_size = step_size_0
+    
+    opt = {}
+    opt['step_size'] = step_size
 
     # initialize iterables
     c_next = c
     fgam, prec = 1,1
     nGrad = _vnorm(grad)
     c0_proj = []
-    opt = {}
     opt['loss'] = [loss]
     # f,ax = plt.subplots()
     # line = ax.plot([0], opt['loss'])
@@ -229,6 +236,7 @@ def optimize( fit:fitting, mode, alg='simple',
         opt['t_total'] = time()-t0 # total optimization time in s
         opt['prec'] = prec # final precision
         opt['MDL'] = fit.MDL(loss)
+        opt['flag_use_peak'] = fit.flag_use_peak
         opt['version'] = __version__
         # opt['fgam'] = fgam # final stepsize modification
         print('\n\tFinished. total time: %.2f min' % ( (time()-t0)/60 ))
@@ -238,8 +246,8 @@ def optimize( fit:fitting, mode, alg='simple',
         # write solution to file
         if name != '':
             name = '_'+name
-        resname = 'opt_%s_%s_nmax%s%s_%s.h5' % (
-            fit.title, modename, fit.ns[-1], name, timestring())
+        resname = 'opt_%s_%s_%s%u%s_%s.h5' % (
+            fit.title, modename, fit.odf_mode, fit.odf_par, name, timestring())
         opt['resname'] = resname
         # os.makedirs('output/', exist_ok=True)
         # with open('output/'+resname, 'wb') as file:
@@ -337,16 +345,16 @@ def backtracking_line_search(
 
 #     return best_alpha, best_f
 
-def line_search(f, x, p, alpha_init=1.0, max_steps=10, tol=1e-6):
+def line_search(function, coefficients, direction, alpha_init=1.0, max_steps=10, tol=1e-6):
     alpha = alpha_init
-    f0 = f(x)
+    f0 = function(coefficients)
     best_alpha = 0.0
     best_f = f0
 
     # Expand phase (increase alpha)
     for _ in range(max_steps):
-        x_new = x + alpha * p
-        f_new = f(x_new)
+        x_new = coefficients + alpha * direction
+        f_new = function(x_new)
         if f_new < best_f:
             best_alpha = alpha
             best_f = f_new
@@ -359,8 +367,8 @@ def line_search(f, x, p, alpha_init=1.0, max_steps=10, tol=1e-6):
     alpha_low = best_alpha
     for _ in range(max_steps):
         alpha = 0.5 * (alpha_low + alpha_high)
-        x_new = x + alpha * p
-        f_new = f(x_new)
+        x_new = coefficients + alpha * direction
+        f_new = function(x_new)
         if f_new < best_f:
             best_alpha = alpha
             best_f = f_new
@@ -373,17 +381,17 @@ def line_search(f, x, p, alpha_init=1.0, max_steps=10, tol=1e-6):
 
     return best_alpha, best_f
 
-def double_line_search(f, x, p, alpha_init=1.0, max_steps=10, tol=1e-6):
+def double_line_search(funciton, coefficients, direction, alpha_init=1.0, max_steps=10, tol=1e-6):
     alpha = alpha_init
-    f0 = f(x)
+    f0 = funciton(coefficients)
     best_alpha = 0.0
     best_f = f0
 
     def expand_contract(alpha, best_alpha, best_f, lam):
         # Expand phase (increase alpha)
         for _ in range(max_steps):
-            x_new = x + alpha * p
-            f_new = f(x_new)
+            x_new = coefficients + alpha * direction
+            f_new = funciton(x_new)
             if f_new < best_f:
                 best_alpha = alpha
                 best_f = f_new
@@ -396,8 +404,8 @@ def double_line_search(f, x, p, alpha_init=1.0, max_steps=10, tol=1e-6):
         alpha_low = best_alpha
         for _ in range(max_steps):
             alpha = 1/lam * (alpha_low + alpha_high)
-            x_new = x + alpha * p
-            f_new = f(x_new)
+            x_new = coefficients + alpha * direction
+            f_new = funciton(x_new)
             if f_new < best_f:
                 best_alpha = alpha
                 best_f = f_new
